package edu.ucla.loni.debabel.processors.keith.calculus;

/**
 * Performs Matrix Operations
 *
 * @author Keith R. Ouellette
 *
 * @version 8 July 2003
 */
public class MatrixManipulator 
{ 
  /** 
   * Returns the dot (scalar) product of two vectors.
   *
   * @param x First vector.
   * @param y Second vector.
   *
   * @return Dot (scalar) product of two vectors.
   *
   * @throws NullPointerException If x or y is null.
   * @throws IllegalArgumentException If x and y do not have same length.
   */
  public static double dot(double x[], double y[])
    {
      if(x == null){
	String msg = "Cannot have a null x.";
	throw new NullPointerException(msg);
      }
      if(y == null){
	String msg = "Cannot have a null y.";
	throw new NullPointerException(msg);
      } 
      if(x.length != y.length) {
	String msg = "Dot product not defined between vectors x and y: " + 
	         "Length mismatch.";
	throw new IllegalArgumentException(msg);
      }
      double z = 0;
      for(int i = 0; i < x.length; i++) {
	z += x[i]*y[i]; 
      }
      return z;
    }

  /** 
   * Returns the sum of two vectors.
   *
   * @param x First vector.
   * @param y Second vector.
   *
   * @return Sum of two vectors.
   *
   * @throws NullPointerException If x or y is null.
   * @throws IllegalArgumentException If x and y do not have same length.
   */
  public static double[] add(double x[], double y[])
    {
      if(x == null){
	String msg = "Cannot have a null x.";
	throw new NullPointerException(msg);
      }
      if(y == null){
	String msg = "Cannot have a null y.";
	throw new NullPointerException(msg);
      } 
      if(x.length != y.length) {
	String msg = "Addition not defined between vectors x and y: " + 
	         "Length mismatch.";
	throw new IllegalArgumentException(msg);
      }
      double z[] = new double[x.length];
      for(int i = 0; i < x.length; i++) {
	z[i] = x[i] + y[i]; 
      }
      return z;
    }

  /** 
   * Returns the result of vector y multiplied by scalar c.
   *
   * @param c Scalar.
   * @param y Second vector.
   *
   * @return Sum of two vectors.
   *
   * @throws NullPointerException If y is null.
   */
  public static double[] scalarMultiply(double c, double y[])
    {
      if(y == null){
	String msg = "Cannot have a null y.";
	throw new NullPointerException(msg);
      } 
      double z[] = new double[y.length];
      for(int i = 0; i < y.length; i++) {
	z[i] = c*y[i]; 
      }
      return z;
    }

  /** 
   * Returns the image of vector x under the transformation represented 
   * by the given matrix. 
   *
   * @param mat Matrix that represents the transformation.
   * @param x Vector to apply the transformation to.
   *
   * @return Image of vector x under the transformation represented 
   * by the given matrix.
   *
   * @throws NullPointerException if mat or x is null.
   * @throws IllegalArgumentException If cannot dot a row of mat with x.
   */
  public static double[] multiply(double mat[][], double x[])
    {
      if(mat == null){
	String msg = "Cannot have a null matrix.";
	throw new NullPointerException(msg);
      }
      if(x == null) {
	String msg = "Cannot have a null x.";
	throw new NullPointerException(msg);
      }
      double[] y = new double[mat.length]; 

      for(int i = 0; i < mat.length; i++) {
	y[i] = dot(mat[i], x);
      }
      return y;
    }
 
  /** 
   * Multiplies two matrices together. If the matrix a has too few columns, 
   * then zeros are added to make multiplication work.
   *
   * @param a First Matrix.
   * @param b Second Matrix.
   *
   * @return The matrix product a*b.
   *
   * @throws NullPointerException If a or b is null.
   * @throws IllegalArgumentException If the multiplication is not defined.
   */
  public static double[][] multiply(double a[][], double b[][])
    {
      if(a == null) {
	String msg = "Cannot have a null matrix A.";
	throw new NullPointerException(msg);
      }
      if(b == null) {
	String msg = "Cannot have a null matrix B.";
	throw new NullPointerException(msg);
      }
      if(a.length == 0) {
	String msg = "Cannot multiply a matrix a with 0 rows.";
	throw new IllegalArgumentException(msg);
      }
      if(b.length == 0) {
	String msg = "Cannot multiply a matrix b with 0 rows.";
	throw new IllegalArgumentException(msg);
      }
      double[][] ab = new double[a.length][]; 

      for(int i = 0; i < a.length; i++) {
	ab[i] = new double[b[0].length];
	for(int j = 0; j < a[i].length; j++ ) {
	  ab[i] = add(ab[i], scalarMultiply(a[i][j], b[j]));	  
	}	
      }
      return ab;
    }

  /** 
   * Computes the trace of a matrix.
   *
   * @param a Matrix to compute the trace of.
   *
   * @return Trace of a.
   *
   * @throws NullPointerException If a is null.
   * @throws IllegalArgumentException If a is not square matrix or has 0 rows.
   * 
   */
  public static double trace(double a[][])
    {
      if(a == null) {
	String msg = "Cannot have a null matrix A.";
	throw new NullPointerException(msg);
      }
      for(int i = 0; i < a.length; i++) {
	if(a[i].length != a.length) {
	  String msg = "Cannot compute trace for a non square matrix.";
	  throw new IllegalArgumentException(msg);
	}
      }
      if(a.length == 0) {
	String msg = "Cannot compute trace for a matrix A with 0 rows.";
	throw new IllegalArgumentException(msg);
      }

      double trace = 0;
      
      for(int i = 0; i < a.length; i++) {
	trace += a[i][i];
      }	
      return trace;
    } 

  /** 
   * Computes the determinant of a matrix.
   *
   * @param a Matrix to compute the determinant of.
   *
   * @return Determinant of a.
   *
   * @throws NullPointerException If a is null.
   * @throws IllegalArgumentException If a is not square matrix or has 0 rows.
   * 
   */
  public static double determinant(double a[][])
    {
      if(a == null) {
	String msg = "Cannot have a null matrix A.";
	throw new NullPointerException(msg);
      }
      for(int i = 0; i < a.length; i++) {
	if(a[i].length != a.length) {
	  String msg = "Cannot compute det for a non square matrix.";
	  throw new IllegalArgumentException(msg);
	}
      }
      if(a.length == 0) {
	String msg = "Cannot compute det for a matrix A with 0 rows.";
	throw new IllegalArgumentException(msg);
      }

      if(a.length == 1) { return a[0][0]; }

      if(a.length == 2) { return a[0][0]*a[1][1] - a[0][1]*a[1][0]; }
      
      double detA = 0;
      int i = 0;

      for(int j = 0; j < a.length; j++) {
	// Expand along first row
	double minorij = determinant(_delete(a, i, j));
	double cofactorij = Math.pow(-1, i + j) * minorij;
	detA += a[i][j] * cofactorij;
      }
      return detA;
    } 

  /** 
   * Returns the transpose of the given matrix.
   */
  public static double[][] transpose(double a[][])
    {
      double t[][] = new double[a[0].length][a.length]; 
      for(int i = 0; i < t.length; i++) {
	for(int j = 0; j < a.length; j++) {
	  t[i][j] = a[j][i];	 
	} 	
      }
      return t;
    }
 
  /** 
   * Returns using Cramer's Rule the preimage of vector b under the 
   * transformation represented by the given matrix . 
   *
   * @param a Matrix that represents the invertible transformation.
   * @param x Vector to apply the transformation to.
   *
   * @return Image of vector x under the transformation represented 
   * by the given matrix.
   *
   * @throws NullPointerException if mat or x is null.
   * @throws IllegalArgumentException If a is not invertible or square.
   */
  public static double[] cramersRule(double a[][], double b[])
    {
      if(b == null) {
	String msg = "Cannot have a null b.";
	throw new NullPointerException(msg);
      }
      
      double detA = determinant(a);

      if(detA == 0){
	String msg = "Matrix a is not invertible. Cannot apply Cramer's Rule"; 
	throw new IllegalArgumentException(msg);
      }
      
      if(a.length != b.length) {
	String msg = "Cannot apply Cramer's Rule: Length mismatch."; 
	throw new IllegalArgumentException(msg);
      }

      double[] x = new double[a.length]; 
    
      for(int j = 0; j < a.length; j++) {
	double aprime[][] = _getCramersMatrix(a,b,j);
	x[j] = 1.0 * determinant(aprime) / detA;
      }
      return x;
    }

  /** 
   * Returns a String representing a 2D array in matrix form.
   *
   * @param a 2D Array which holds the matrix.
   *
   * @return String representing a 2D array in matrix form.
   */
  public static String print(double a[][])
    {
      String result = "";
      for(int i = 0; i < a.length; i++) {
	result += print(a[i]) + "\n";
      }
      return result;
    }

  /** 
   * Returns a String representing a 1D array in matrix form.
   *
   * @param a 1D Array which holds the matrix.
   *
   * @return String representing a 1D array in matrix form.
   */
  public static String print(double a[])
    {
      String result = "[";
      for(int j = 0; j < a.length; j++) {
	  result += a[j] + ", ";
      }
      if(result.endsWith(", ")) {
	result = result.substring(0, result.length() - 2);
      }
      result += "]";
      return result;
    }

  public static void main(String args[]) 
    {
     double a[][] = { {2,1,-1}, {3,1,2}, {0,-2,-3} };
     double b[][] = { {3,2}, {-1,1}, {6,4} };
     double c[][] = { {2,1,0}, {-1,3,2} };
     double d[][] = { {1,-1,1}, {2,-1,2}, {3,0,3} };
 
     System.out.println(print(multiply(a,d)) + "\n" 
             + print(transpose(multiply(a,d))) + "\n");
     System.out.println(print(multiply(d,a)) + "\n");
     System.out.println(print(multiply(b,c)) + "\n");
     System.out.println(print(multiply(c,b)) + "\n");
     System.out.println(print(multiply(a,b)) + "\n");
     System.out.println(print(multiply(b,a)) + "\n");

     double e[][] = {{2,-1,3,0},{4,-2,7,0},{-3,-4,1,5},{6,-6,8,0}};
     System.out.println("det(E) = " + determinant(e) + "\n");

     double f[][] = {{1,2,-1},{2,1,0},{-1,1,2}};
     double b0[] = {2,3,4};
     System.out.println("Solutions: x = " + print(cramersRule(f, b0)));
     System.out.println("x1 = 7/9 " + (7.0/9));
     System.out.println("x2 = 13/9 " + (13.0/9));
     System.out.println("x3 = 5/3 " + (5.0/3));
   } 

  /** 
   * Returns the transpose of the matrix used in the implementation of 
   * Cramer's rule. (Replaces specified row of a transpose with b).
   */
  private static double[][] _getCramersMatrix(double a[][], double b[],int row)
    {
      double m[][] = new double[a.length][a.length]; 
      for(int i = 0; i < a.length; i++) {
	for(int j = 0; j < a.length; j++) { 	  
	  if(i == row) { m[i][j] = b[j]; }
	  else { m[i][j] = a[j][i]; } 	  
	}
      }
      return m;
    } 

  /** 
   * Returns the matrix obtained by deleting the given row and column.
   *
   * @param a Matric to remove the row and column from.
   * @param row Row index to remove.
   * @param col Column index to remove.
   *
   * @return Matrix obtained by deleting the given row and column.
   *
   * @throws NullPointerException If a is null.
   * @throws IllegalArgumentException If row or col is invalid.
   */
  private static double[][] _delete(double a[][], int row, int col)
    {
      if(a == null) {
	String msg = "Cannot have a null matrix A.";
	throw new NullPointerException(msg);
      }
      for(int i = 0; i < a.length; i++) {
	if(a[i].length != a.length) {
	  String msg = "Matrix must be square for delete operation.";
	  throw new IllegalArgumentException(msg);
	}
      }
      if(a.length == 0) {
	String msg = "Cannot delete a matrix A with 0 rows.";
	throw new IllegalArgumentException(msg);
      }
      if(row >= a.length || row < 0) {
	String msg = "Invalid row index.";
	throw new IllegalArgumentException(msg);
      }
      if(col >= a.length || col < 0) {
	String msg = "Invalid col index.";
	throw new IllegalArgumentException(msg);
      }

      double m[][] = new double[a.length - 1][a.length - 1];
      
      for(int i = 0; i < a.length; i++) {
	if(i == row) { continue; }
	for(int j = 0; j < a.length; j++) {	  
	  if(j == col) { continue; }
	  
	  int r = (i <= row ? i : i-1);
	  int c = (j <= col ? j : j-1);
	  
	  m[r][c] = a[i][j];	    
	}
      }
      return m;
    }
}
