/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.keith.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Stateful Processor that sorts DICOM images.
 *
 * @version 15 December 2003
 */
public class DicomImageSorter extends Processor implements Stateful
{
  /** Description of this Processor. */
  private static String DESC = "Sorts DICOM images.";

  /** Input Node name for the X coordinate of the first image pixel. */
  private static String X_NAME = "Image Position (X)";

  /** Input Node description for the X coordinate of the first image pixel. */
  private static String X_DESC = "X coordinate of the first image pixel.";

  /** Input Node name for the Y coordinate of the first image pixel. */
  private static String Y_NAME = "Image Position (Y)";

  /** Input Node description for the Y coordinate of the first image pixel. */
  private static String Y_DESC = "Y coordinate of the first image pixel.";

  /** Input Node name for the Z coordinate of the first image pixel. */
  private static String Z_NAME = "Image Position (Z)";

  /** Input Node description for the Z coordinate of the first image pixel. */
  private static String Z_DESC = "Z coordinate of the first image pixel.";

  /** Input Node name for the X coordinate of the sorting direction cosine. */
  private static String SORT_X_NAME = "Sorting Direction Cosine (X)";

  /** Input Node desc for the X coord of the sorting direction cosine. */
  private static String SORT_X_DESC =
  "Direction cosine (along X) of the vector to sort the images along.";

  /** Input Node name for the Y coordinate of the sorting direction cosine. */
  private static String SORT_Y_NAME = "Sorting Direction Cosine (Y)";

  /** Input Node desc for the Y coord of the sorting direction cosine. */
  private static String SORT_Y_DESC =
  "Direction cosine (along Y) of the vector to sort the images along.";

  /** Input Node name for the Z coordinate of the sorting direction cosine. */
  private static String SORT_Z_NAME = "Sorting Direction Cosine (Z)";

  /** Input Node desc for the Z coord of the sorting direction cosine. */
  private static String SORT_Z_DESC =
  "Direction cosine (along Z) of the vector to sort the images along.";

  /** Input Node Name for the image. */
  private static String IMAGE_NAME = "Image";
  
  /** Input Node Description for the image. */
  private static String IMAGE_DESC = "Image to sort.";

  /** Output Node Name for the sorted image array. */
  private static String ARRAY_NAME = "Sorted Image Array.";

  /** Output Node Description for the sorted image array. */
  private static String ARRAY_DESC = "Images sorted by DICOM field.";

  /** Output Node Name for the average slice spacing. */
  private static String SPACING_NAME = "Average Slice Spacing";

  /** Output Node Description for the average slice spacing. */
  private static String SPACING_DESC = "Average spacing between slices.";

  /** Output Node name for the X coordinate of the first sorted image pixel. */
  private static String ORIGIN_X_NAME = "Origin (X)";

  /** Output Node desc for the X coordinate of the first sorted image pixel. */
  private static String ORIGIN_X_DESC =
  "X coordinate of the first sorted image's pixel.";

  /** Output Node name for the Y coordinate of the first sorted image pixel. */
  private static String ORIGIN_Y_NAME = "Origin (Y)";

  /** Output Node desc for the Y coordinate of the first sorted image pixel. */
  private static String ORIGIN_Y_DESC =
  "Y coordinate of the first sorted image's pixel.";

  /** Output Node name for the Z coordinate of the first sorted image pixel. */
  private static String ORIGIN_Z_NAME = "Origin (Z)";

  /** Output Node desc for the Z coordinate of the first sorted image pixel. */
  private static String ORIGIN_Z_DESC =
  "Z coordinate of the first sorted image's pixel.";

  /** Sorted Dicom Images. */
  private Vector _sortedImages;

  /** Coordinate to sort along. */
  private Double[] _sortingCoordinate;

  /** Dicom Image whose top left coordinates represent the origin. */
  private DicomImage _originImage;

  /**
   * Constructs a Dicom Image Sorter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomImageSorter(String name)
    {
      super(name, DESC);

      // Create the Input Node for the X coordinate of the first image pixel
      Class type = Double.class;
      InputNode inputNode = new InputNode(X_NAME, X_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the Y coordinate of the first image pixel
      type = Double.class;
      inputNode = new InputNode(Y_NAME, Y_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the Z coordinate of the first image pixel
      type = Double.class;
      inputNode = new InputNode(Z_NAME, Z_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for x of the coordinate to sort along
      type = Double.class;
      inputNode = new InputNode(SORT_X_NAME, SORT_X_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for x of the coordinate to sort along
      type = Double.class;
      inputNode = new InputNode(SORT_Y_NAME, SORT_Y_DESC, type);
      _addInputNode(inputNode);
      
      // Create the Input Node for x of the coordinate to sort along
      type = Double.class;
      inputNode = new InputNode(SORT_Z_NAME, SORT_Z_DESC, type);
      _addInputNode(inputNode);
      
      // Create the Input Node for the image
      type = BufferedImage.class;
      inputNode = new InputNode(IMAGE_NAME, IMAGE_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the sorted image array
      type = BufferedImage[].class;
      OutputNode node = new OutputNode(ARRAY_NAME, ARRAY_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the average slice spacing
      type = Double.class;
      node = new OutputNode(SPACING_NAME, SPACING_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the X coordinate of the origin
      type = Double.class;
      node = new OutputNode(ORIGIN_X_NAME, ORIGIN_X_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the Y coordinate of the origin
      type = Double.class;
      node = new OutputNode(ORIGIN_Y_NAME, ORIGIN_Y_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the Z coordinate of the origin
      type = Double.class;
      node = new OutputNode(ORIGIN_Z_NAME, ORIGIN_Z_DESC, type);
      _addOutputNode(node);

      // Clear the sorted images
      reinitializeState();
    }

  /**
   * Constructs a Dicom Image Sorter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomImageSorter(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Input values
	Double x = (Double)_getInputValue(X_NAME, null);
	Double y = (Double)_getInputValue(Y_NAME, null);
	Double z = (Double)_getInputValue(Z_NAME, null);

	// Get the sorting coordinate
	_sortingCoordinate = new Double[3];
	_sortingCoordinate[0] = (Double)_getInputValue(SORT_X_NAME, null);
	_sortingCoordinate[1] = (Double)_getInputValue(SORT_Y_NAME, null);
	_sortingCoordinate[2] = (Double)_getInputValue(SORT_Z_NAME, null);

	if (_sortingCoordinate[0] == null) {
	  String msg = "No X coordinate to sort along.";
	  throw new ProcessingException(msg);
	}
	if (_sortingCoordinate[1] == null) {
	  String msg = "No Y coordinate to sort along.";
	  throw new ProcessingException(msg);
	}
	if (_sortingCoordinate[2] == null) {
	  String msg = "No Z coordinate to sort along.";
	  throw new ProcessingException(msg);
	}

	BufferedImage image = (BufferedImage)_getInputValue(IMAGE_NAME, null);
	if (image == null) {
	  String msg = "No image to sort.";
	  throw new ProcessingException(msg);
	}

	// Construct a Dicom Image from the Input values	
	DicomImage dicomImage = new DicomImage(x, y, z, image);
	if(_sortedImages.isEmpty()) { _originImage = dicomImage; }
	_sortedImages.add(dicomImage);

	// Sort the Dicom Images
	Collections.sort(_sortedImages, _getComparator());

	// Set the array of sorted Buffered Images
	BufferedImage[] images = new BufferedImage[ _sortedImages.size() ];
	for (int i = 0; i < _sortedImages.size(); i++) {
	  dicomImage = (DicomImage)_sortedImages.elementAt(i);
	  images[i] = dicomImage.getBufferedImage();
	}
	_setOutputValue(ARRAY_NAME, images);

	// Set the average slice spacing
	double sliceSpacing = _computeAverageSliceSpacing(_sortedImages);
	_setOutputValue(SPACING_NAME, new Double(sliceSpacing));

	// Set the origin
	if ( _sortedImages.size() > 0 ) {

	  // Set the coordinates of the first sorted image
	  dicomImage = (DicomImage)_sortedImages.elementAt(0);
	  _setOutputValue(ORIGIN_X_NAME, dicomImage.getX());
	  _setOutputValue(ORIGIN_Y_NAME, dicomImage.getY());
	  _setOutputValue(ORIGIN_Z_NAME, dicomImage.getZ());
	}
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Reinitializes the state.  This is called by a Loop Module before a loop is
   * started.
   */
  public void reinitializeState()
    {
      _sortedImages = new Vector();
      _originImage = null;
    }

  /**
   * Computes the average slice spacing between the ordered DICOM images.
   *
   * @param sortedImages Sorted DICOM images.
   *
   * @return Average slice spacing between the ordered DICOM images.
   */
  private double _computeAverageSliceSpacing(Vector sortedImages)
    {
      double sliceSpacing = 0;

      // Get the average slice spacing
      int numberOfImages = sortedImages.size();
      for (int i = 0; i < numberOfImages-1; i++) {
	DicomImage d1 = (DicomImage)sortedImages.elementAt(i);
	DicomImage d2 = (DicomImage)sortedImages.elementAt(i+1);

	// Determine if the Image Position coordinates exists
	if ( d1.getX() != null && d1.getY() != null &&
	     d1.getZ() != null && d2.getX() != null && 
	     d2.getY() != null && d2.getZ() != null )
	  {
	    double x1 = d1.getX().doubleValue();
	    double x2 = d2.getX().doubleValue();
	    double y1 = d1.getY().doubleValue();
	    double y2 = d2.getY().doubleValue();
	    double z1 = d1.getZ().doubleValue();
	    double z2 = d2.getZ().doubleValue();

	    double dsq = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);
	    sliceSpacing += Math.sqrt(dsq);
	  }
      }

      // Return the average slice spacing
      int numberOfSpaces = numberOfImages - 1;
      if (numberOfSpaces > 0) { sliceSpacing /= numberOfSpaces; }
      return sliceSpacing;
    }

  /**
   * Gets a Comparator for sorting DICOM images.
   *
   * @return Comparator for sorting DICOM images.
   */
  private Comparator _getComparator()
    {
      return new Comparator()
	{	  
	  /**
	   * Compares its two arguments for order.
	   *
	   * @param o1 The first object to be compared.
	   * @param o2 The second object to be compared.
	   *
	   * @return A negative integer, zero, or a positive integer as the
	   *         first argument is less than, equal to, or greater than
	   *         the second.
	   *
	   * @throws IllegalArgumentException If there isn't enough information
	   *                                  to compare them.
	   */
	  public int compare(Object o1, Object o2)
	    {
	      // Determine if the Image Position coordinates exist
	      try { 
		double m1 = _projectionMagnitude((DicomImage)o1);
		double m2 = _projectionMagnitude((DicomImage)o2);
	      
		// *****
		//
		// DICOM STANDARDS:
		//
		// The x-y-z coordinate system on a patient is defined so
		// the x axis points to the patient's left, the y axis points
		// to the patient's posterior, and the z axis points to the 
		// patient's head:
		//
		//    (*+*)     ^ Z
		//     _|_      |
		//      |       | Y
		//     / \      x -- > X
		// *****
		
		// Order along increasing coordinates
		if (m1 < m2) { return -1; }
		if (m1 == m2) { return 0; }
		return 1;
	      }
	      catch (Exception e) {
		// Can't make the comparison
		throw new IllegalArgumentException("Not enough information to "
                    + "sort DICOM images.");
	      }
	    }
      
	  /**
	   * Indicates whether some object is equal to this Comparator.
	   *
	   * @param obj The reference object with which to compare.
	   *
	   * @return True only if the specified object is also a comparator
	   *         and it imposes the same ordering as this comparator.
	   */
	  public boolean equals(Object obj)
	    {
	      return super.equals(obj);
	    }

	  /** 
	   * Computes the magnitude of the projection of the given image's
	   * top left coordinate onto the sorting directional vector. 
	   *
	   * @param d DicomImage whose coordinates are to be projected.
	   *
	   * @return Magnitude of the projection of the given image's
	   * top left coordinate onto the sorting directional vector.
	   */
	  private double _projectionMagnitude(DicomImage d)
	    {
	      // Get the coordinates of the DicomImage d
	      double vx = d.getX().doubleValue();
	      double vy = d.getY().doubleValue();
	      double vz = d.getZ().doubleValue();
	      
	      // Get the origin (first image of the loop)
	      // Not necessarily the first image in sorted order
	      double ox = _originImage.getX().doubleValue();
	      double oy = _originImage.getY().doubleValue();
	      double oz = _originImage.getZ().doubleValue();
	      
	      // Get the displacement of image d w.r.t origin
	      double ovx = vx - ox;
	      double ovy = vy - oy;
	      double ovz = vz - oz;
	      
	      // Get the sorting unit vector
	      double ux = _sortingCoordinate[0].doubleValue();
	      double uy = _sortingCoordinate[1].doubleValue();
	      double uz = _sortingCoordinate[2].doubleValue();
	      
	      // Dot v with u to get the projection onto u
	      return ovx*ux + ovy*uy + ovz*uz;  
	    }
	};
    }

  /** Class used to hold a Buffered Image with its metadata. */
  private class DicomImage
  {
    /** X coordinate of the first image pixel. */
    private Double _x;

    /** Y coordinate of the first image pixel. */
    private Double _y;

    /** Z coordinate of the first image pixel. */
    private Double _z;

    /** Buffered Image. */
    private BufferedImage _image;

    /**
     * Constructs a DicomImage.
     *
     * @param x X coordinate of the first image pixel.
     * @param y Y coordinate of the first image pixel.
     * @param z Z coordinate of the first image pixel.
     * @param image Buffered Image.
     */
    public DicomImage(Double x, Double y, Double z, BufferedImage image)
      {
	_x = x;
	_y = y;
	_z = z;
	_image = image;
      }

    /**
     * Determines if valid X,Y,Z coordinates are available.
     *
     * @return True if valid X,Y,Z coordinates are available; false otherwise.
     */
    public boolean hasValidCoordinates()
      {
	return (_x != null && _y != null && _z != null);
      }

    /**
     * Gets the X coordinate.
     *
     * @return X coordinate of the first image pixel.
     */
    public Double getX()
      {
	return _x;
      }

    /**
     * Gets the Y coordinate.
     *
     * @return Y coordinate of the first image pixel.
     */
    public Double getY()
      {
	return _y;
      }

    /**
     * Gets the Z coordinate.
     *
     * @return Z coordinate of the first image pixel.
     */
    public Double getZ()
      {
	return _z;
      }

    /**
     * Gets the Buffered Image.
     *
     * @return Buffered Image.
     */
    public BufferedImage getBufferedImage()
      {
	return _image;
      }
  }
}
