package edu.ucla.loni.debabel.processors.keith.talairach;

import edu.ucla.loni.debabel.processors.keith.calculus.MatrixManipulator;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.awt.image.WritableRaster;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Processor that reformats a set of Buffered Images in Talairach space.
 *
 * @author Keith R. Ouellette
 *
 * @version 14 July 2003
 */
public class RasterRotator extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Reformats images in Talairach space.";
  
  /** Node name for the old row x direction cosine. */
  private static String ROW_X_COS_NAME = "Old Row x Direction Cosine";
  
  /** Node description for the old row x direction cosine Node. */
  private static String ROW_X_COS_DESC = "Old Row x Direction Cosine";
  
  /** Node name for the old row y direction cosine. */
  private static String ROW_Y_COS_NAME = "Old Row y Direction Cosine";
  
  /** Node description for the old row y direction cosine Node. */
  private static String ROW_Y_COS_DESC = "Old Row y Direction Cosine";
  
  /** Node name for the old row z direction cosine.*/
  private static String ROW_Z_COS_NAME = "Old Row z Direction Cosine";
  
  /** Node description for the old row z direction cosine Node. */
  private static String ROW_Z_COS_DESC = "Old Row z Direction Cosine";

  /** Node name for the old column x direction cosine. */
  private static String COL_X_COS_NAME = "Old Col x Direction Cosine";
  
  /** Node description for the old column x direction cosine Node. */
  private static String COL_X_COS_DESC = "Old Col x Direction Cosine";
  
  /** Node name for the old column y direction cosine. */
  private static String COL_Y_COS_NAME = "Old Col y Direction Cosine";
  
  /** Node description for the old column y direction cosine Node. */
  private static String COL_Y_COS_DESC = "Old Col y Direction Cosine";
  
  /** Node name for the old column z direction cosine.*/
  private static String COL_Z_COS_NAME = "Old Col z Direction Cosine";
  
  /** Node description for the old column z direction cosine Node. */
  private static String COL_Z_COS_DESC = "Old Col z Direction Cosine";

  /** Node name for the old stack x direction cosine. */
  private static String STACK_X_COS_NAME = "Old Stack x Direction Cosine";
  
  /** Node description for the old stack x direction cosine Node. */
  private static String STACK_X_COS_DESC = "Old Stack x Direction Cosine";
  
  /** Node name for the old stack y direction cosine. */
  private static String STACK_Y_COS_NAME = "Old Stack y Direction Cosine";
  
  /** Node description for the old stack y direction cosine Node. */
  private static String STACK_Y_COS_DESC = "Old Stack y Direction Cosine";
  
  /** Node name for the old stack z direction cosine.*/
  private static String STACK_Z_COS_NAME = "Old Stack z Direction Cosine";
  
  /** Node description for the old stack z direction cosine Node. */
  private static String STACK_Z_COS_DESC = "Old Stack z Direction Cosine";

  /** Node name for the new row x direction cosine. */
  private static String NEW_ROW_X_COS_NAME = "New Row x Direction Cosine";
 
  /** Node description for the new row x direction cosine Node. */
  private static String NEW_ROW_X_COS_DESC = "New Row x Direction Cosine";
 
  /** Node name for the new row y direction cosine. */
  private static String NEW_ROW_Y_COS_NAME = "New Row y Direction Cosine";

  /** Node description for the new row y direction cosine Node. */
  private static String NEW_ROW_Y_COS_DESC = "New Row y Direction Cosine";
  
  /** Node name for the new row z direction cosine.*/
  private static String NEW_ROW_Z_COS_NAME = "New Row z Direction Cosine";

  /** Node description for the new row z direction cosine Node. */
  private static String NEW_ROW_Z_COS_DESC = "New Row z Direction Cosine";

  /** Node name for the new column x direction cosine. */
  private static String NEW_COL_X_COS_NAME = "New Col x Direction Cosine";

  /** Node description for the new column x direction cosine Node. */
  private static String NEW_COL_X_COS_DESC = "New Col x Direction Cosine";

  /** Node name for the new column y direction cosine. */
  private static String NEW_COL_Y_COS_NAME = "New Col y Direction Cosine";
 
  /** Node description for the new column y direction cosine Node. */
  private static String NEW_COL_Y_COS_DESC = "New Col y Direction Cosine";
 
  /** Node name for the new column z direction cosine.*/
  private static String NEW_COL_Z_COS_NAME = "New Col z Direction Cosine";
 
  /** Node description for the new column z direction cosine Node. */
  private static String NEW_COL_Z_COS_DESC = "New Col z Direction Cosine";

  /** Node name for the new stack x direction cosine. */
  private static String NEW_STACK_X_COS_NAME = "New Stack x Direction Cosine";

  /** Node description for the new stack x direction cosine Node. */
  private static String NEW_STACK_X_COS_DESC = "New Stack x Direction Cosine";

  /** Node name for the new stack y direction cosine. */
  private static String NEW_STACK_Y_COS_NAME = "New Stack y Direction Cosine";
 
  /** Node description for the new stack y direction cosine Node. */
  private static String NEW_STACK_Y_COS_DESC = "New Stack y Direction Cosine";
  
  /** Node name for the new stack z direction cosine.*/
  private static String NEW_STACK_Z_COS_NAME = "New Stack z Direction Cosine";
  
  /** Node description for the new stack z direction cosine Node. */
  private static String NEW_STACK_Z_COS_DESC = "New Stack z Direction Cosine";

  /** Node name for the row step size. */
  private static String ROW_STEP_NAME = "Row Spacing";

  /** Node description for the row step size. */
  private static String ROW_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the rows.";
  /** Node name for the column step size. */
  private static String COL_STEP_NAME = "Column Spacing";

  /** Node description for the column step size. */
  private static String COL_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the columns.";
  /** Node name for the stack step size. */
  private static String STACK_STEP_NAME = "Stack Spacing";

  /** Node description for the stack step size. */
  private static String STACK_STEP_DESC = "Step size between adjacent " +
                                          "voxels in the stack.";
  /** Node name for the images. */
  private static String IMAGES_NAME = "Images";

  /** Node description for the images. */
  private static String IMAGES_DESC = "Images.";
 
  /**
   * Constructs a RasterRotator.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public RasterRotator(String name)
    {
      super(name, DESC);

      // Create the Input Nodes for the current Talairach Coordinates
      Class type = Double.class;
      InputNode inputNode = new InputNode(ROW_X_COS_NAME, ROW_X_COS_DESC,type);
      _addInputNode(inputNode);
      
      type = Double.class;
      inputNode = new InputNode(ROW_Y_COS_NAME, ROW_Y_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(ROW_Z_COS_NAME, ROW_Z_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(COL_X_COS_NAME, COL_X_COS_DESC, type);
      _addInputNode(inputNode); 
      
      type = Double.class;
      inputNode = new InputNode(COL_Y_COS_NAME, COL_Y_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(COL_Z_COS_NAME, COL_Z_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(STACK_X_COS_NAME, STACK_X_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(STACK_Y_COS_NAME, STACK_Y_COS_DESC, type);
      _addInputNode(inputNode);
      
      type = Double.class;
      inputNode = new InputNode(STACK_Z_COS_NAME, STACK_Z_COS_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Nodes for the new Talairach Coordinates
      type = Double.class;
      inputNode = new InputNode(NEW_ROW_X_COS_NAME, NEW_ROW_X_COS_DESC,type);
      _addInputNode(inputNode);
      
      type = Double.class;
      inputNode = new InputNode(NEW_ROW_Y_COS_NAME, NEW_ROW_Y_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(NEW_ROW_Z_COS_NAME, NEW_ROW_Z_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(NEW_COL_X_COS_NAME, NEW_COL_X_COS_DESC, type);
      _addInputNode(inputNode); 
      
      type = Double.class;
      inputNode = new InputNode(NEW_COL_Y_COS_NAME, NEW_COL_Y_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(NEW_COL_Z_COS_NAME, NEW_COL_Z_COS_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(NEW_STACK_X_COS_NAME, NEW_STACK_X_COS_DESC, 
				type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(NEW_STACK_Y_COS_NAME, NEW_STACK_Y_COS_DESC,
				 type);
      _addInputNode(inputNode);
      
      type = Double.class;
      inputNode = new InputNode(NEW_STACK_Z_COS_NAME, NEW_STACK_Z_COS_DESC,
				type);
      _addInputNode(inputNode);

      // Create the Input Nodes for the spacings
      type = Double.class;
      inputNode = new InputNode(ROW_STEP_NAME, ROW_STEP_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(COL_STEP_NAME, COL_STEP_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(STACK_STEP_NAME, STACK_STEP_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the images
      type = BufferedImage[].class;
      inputNode = new InputNode(IMAGES_NAME, IMAGES_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Nodes for the spacings
      type = Double.class;
      OutputNode node = new OutputNode("New "+ROW_STEP_NAME,
				       "New "+ROW_STEP_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode("New "+COL_STEP_NAME, "New "+ COL_STEP_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode("New "+STACK_STEP_NAME, "New "+ STACK_STEP_DESC,
			    type);
      _addOutputNode(node);

      // Create the Output Node for the images
      type = BufferedImage[].class;
      node = new OutputNode("New " + IMAGES_NAME, "New " + IMAGES_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Constructs a RasterRotator using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public RasterRotator(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the Input Node values for the current Talairach Coordinates
	Double rowXCos = null;
	Object value = _getInputValue(ROW_X_COS_NAME, null);
	if (value instanceof Double) { rowXCos = (Double)value; }

	Double rowYCos = null;
	value = _getInputValue(ROW_Y_COS_NAME, null);
	if (value instanceof Double) { rowYCos = (Double)value; }

	Double rowZCos = null;
	value = _getInputValue(ROW_Z_COS_NAME, null);
	if (value instanceof Double) { rowZCos = (Double)value; }

	Double colXCos = null;
	value = _getInputValue(COL_X_COS_NAME, null);
	if (value instanceof Double) { colXCos = (Double)value; }

	Double colYCos = null;
	value = _getInputValue(COL_Y_COS_NAME, null);
	if (value instanceof Double) { colYCos = (Double)value; }

	Double colZCos = null;
	value = _getInputValue(COL_Z_COS_NAME, null);
	if (value instanceof Double) { colZCos = (Double)value; }

	Double stackXCos = null;
	value = _getInputValue(STACK_X_COS_NAME, null);
	if (value instanceof Double) { stackXCos = (Double)value; }

	Double stackYCos = null;
	value = _getInputValue(STACK_Y_COS_NAME, null);
	if (value instanceof Double) { stackYCos = (Double)value; }

	Double stackZCos = null;
	value = _getInputValue(STACK_Z_COS_NAME, null);
	if (value instanceof Double) { stackZCos = (Double)value; }

	// Get the Input Node values for the new Talairach Coordinates
	Double newRowXCos = null;
	value = _getInputValue(NEW_ROW_X_COS_NAME, null);
	if (value instanceof Double) { newRowXCos = (Double)value; }

	Double newRowYCos = null;
	value = _getInputValue(NEW_ROW_Y_COS_NAME, null);
	if (value instanceof Double) { newRowYCos = (Double)value; }

	Double newRowZCos = null;
	value = _getInputValue(NEW_ROW_Z_COS_NAME, null);
	if (value instanceof Double) { newRowZCos = (Double)value; }

	Double newColXCos = null;
	value = _getInputValue(NEW_COL_X_COS_NAME, null);
	if (value instanceof Double) { newColXCos = (Double)value; }

	Double newColYCos = null;
	value = _getInputValue(NEW_COL_Y_COS_NAME, null);
	if (value instanceof Double) { newColYCos = (Double)value; }

	Double newColZCos = null;
	value = _getInputValue(NEW_COL_Z_COS_NAME, null);
	if (value instanceof Double) { newColZCos = (Double)value; }

	Double newStackXCos = null;
	value = _getInputValue(NEW_STACK_X_COS_NAME, null);
	if (value instanceof Double) { newStackXCos = (Double)value; }

	Double newStackYCos = null;
	value = _getInputValue(NEW_STACK_Y_COS_NAME, null);
	if (value instanceof Double) { newStackYCos = (Double)value; }

	Double newStackZCos = null;
	value = _getInputValue(NEW_STACK_Z_COS_NAME, null);
	if (value instanceof Double) { newStackZCos = (Double)value; }
       
	// Get the Input Node values for the spacings
	Double rSpacing = null;
	value = _getInputValue(ROW_STEP_NAME, null);
	if (value instanceof Double) { rSpacing = (Double)value; }

	Double cSpacing = null;
	value = _getInputValue(COL_STEP_NAME, null);
	if (value instanceof Double) { cSpacing = (Double)value; }

	Double sSpacing = null;
	value = _getInputValue(STACK_STEP_NAME, null);
	if (value instanceof Double) { sSpacing = (Double)value; }

	// Get the Input Node value for the images
	BufferedImage[] images = null;
	value = _getInputValue(IMAGES_NAME, null);
	if (value instanceof BufferedImage[]) {images=(BufferedImage[])value;}

	// Determine if enough information is provided
	BufferedImage[] newImages = null;
	if ( rowXCos != null && rowYCos != null && rowZCos != null &&
	     colXCos != null && colYCos != null && colZCos != null &&
	     stackXCos != null && stackYCos != null && stackZCos != null &&
	     newRowXCos != null && newRowYCos != null && newRowZCos != null &&
	     newColXCos != null && newColYCos != null && newColZCos != null &&
	     newStackXCos != null && newStackYCos != null 
	     && newStackZCos != null && images != null)
	  {
	    // Determine the transformation between old and new Talairach 
	    // Coordinates.
	    double a[][] = new double[][] {
	      { rowXCos.doubleValue(), colXCos.doubleValue(),
		  stackXCos.doubleValue() }, 
	      { rowYCos.doubleValue(), colYCos.doubleValue(),
		  stackYCos.doubleValue() },
	      { rowZCos.doubleValue(), colZCos.doubleValue(), 
		  stackZCos.doubleValue() }};
	    
	    double b[][] = new double[][] {
	      { newRowXCos.doubleValue(), newColXCos.doubleValue(), 
		  newStackXCos.doubleValue() }, 
	      { newRowYCos.doubleValue(), newColYCos.doubleValue(), 
		    newStackYCos.doubleValue() },
	      { newRowZCos.doubleValue(), newColZCos.doubleValue(), 
		      newStackZCos.doubleValue() } };

	    // Standard basis vectors
	    double e[][] = {{1,0,0}, {0,1,0}, {0,0,1}};
	    double[][] bainv = _createChangeBasisMatrix(a, b);
	    
	    // Determine X', and Z' in XYZ coordinates	   
	    double xprime[] = MatrixManipulator.multiply(bainv, e[0]);
	    double zprime[] = MatrixManipulator.multiply(bainv, e[2]);

	    // Project the Z' coordinate axis onto XY plane
	    double pzprime[] = {zprime[0], zprime[1], 0};
 
	    // Determine the rotation angle to align projZ' with X axis
	    double costheta = MatrixManipulator.dot(e[0], pzprime);
	    // Determine the rotation angle to align Z' with Z (new axis)
	    double cosphi = MatrixManipulator.dot(e[2], pzprime);
  
	    // Recover the angles from their respective cosine values.
	    double theta = Math.acos(costheta);
	    double phi = Math.acos(cosphi);
	    double rho = _getRho(xprime, e, theta, phi);

	    // Use AffineTransform to rotate slices by that angle 
	    Vector spacings = new Vector();
	    images = _scaleRotateAndRescale(images,theta, spacings);
	   
	    // Spacing dimension check....
	    WritableRaster raster = images[0].getRaster();
	    int width = images[0].getWidth();
	    int height = images[0].getHeight();
	    rSpacing = ((Double) spacings.elementAt(0));
	    cSpacing = ((Double) spacings.elementAt(1));

	    // Resample along perpendicular plane (XZ plane) 
	    images = _reformat(images);
	    Double temp = cSpacing;
	    sSpacing = cSpacing;
	    cSpacing = temp;

	    // Reverse the order of the images
	    images = _reverseOrder(images);
	    // Right Eye becomes left eye so must flip!!!!!!!
	    
	    // Use AffineTransform to rotate slices by that angle 
	    spacings.clear();
	    images = _scaleRotateAndRescale(images, phi, spacings);
	       
	    // Spacing dimension check.... 
	    rSpacing = ((Double) spacings.elementAt(0));
	    sSpacing = ((Double) spacings.elementAt(1));

	    // Reverse the order of the images
	    images = _reverseOrder(images);
	    // Left Eye becomes right eye again so OK !!!!!!! 
	    // Resample again to XY plane
	    images = _reformat(images);	
	    temp = cSpacing;
	    sSpacing = cSpacing;
	    cSpacing = temp;
	    
	    // Use AffineTransform to rotate slices by that angle 
	    spacings.clear();
	    images = _scaleRotateAndRescale(images, rho, spacings);

	    // Spacing dimension check.... 
	    rSpacing = ((Double) spacings.elementAt(0));
	    cSpacing = ((Double) spacings.elementAt(1)); 
	  }
	// Set the output values
	_setOutputValue("New "+ ROW_STEP_NAME, rSpacing);
	_setOutputValue("New "+ COL_STEP_NAME, cSpacing);
	_setOutputValue("New "+ STACK_STEP_NAME, sSpacing);
	_setOutputValue("New "+ IMAGES_NAME, images);
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
 
  /** 
   * Creates the matrix B*Inv(A), which, when B is invertible, is the change
   * of basis matrix from the column vectors of A to the column vectors of B.
   * 
   * @param a Matrix whose colmuns comprise the old basis.
   * @param b Matrix whose colmuns comprise the new basis.
   *
   * @return The matrix B*Inv(A).
   */
  private double[][] _createChangeBasisMatrix(double a[][], double b[][])
    {
      double e[][] = {{1,0,0}, {0,1,0}, {0,0,1}};
      double ainvt[][] = new double[a.length][];
      
      // Construct the transpose of inverse transformation repres. by a
      for(int i = 0; i < a.length; i++) {
	ainvt[i] = MatrixManipulator.cramersRule(a, e[i]);
      }
      // Undo the transpose 
      double ainv[][] = MatrixManipulator.transpose(ainvt);
      
      // Construct the change of basis matrix from old cosines to new 
      // with respect to the standard basis.
      return MatrixManipulator.multiply(b,ainv);   
    }

  /** 
   * Computes the third angle to rotate XY slices so that X' lines up with 
   * X and Y' lines up with Y. Part of the total AffineTransformation that
   * acts on the image so that it conforms to the new directional cosines. 
   *
   * @param xp The unit basis Vector X' (Old row basis vector)
   * @param e Matrix whose rows are target basis vectors X, Y, Z
   * written in terms of the ordered basis {X, Y, Z} (ie identity matrix).
   * @param theta Rotation angle in XY plane to align projZ' with X axis.
   * @param phi Rotation angle in XZ plane to align Z' with Z (new axis).
   * 
   * @return Angle (in radians) to rotate XY slices so that X' lines up with 
   * X and Y' lines up with Y.
   */
  private double _getRho(double xp[], double e[][], double theta, double phi)
    {
      // Rotate X' around Z axis
      Point2D.Double pxp = new Point2D.Double(xp[0], xp[1]);
      Point2D.Double pxpdest1 = new Point2D.Double();
      Point2D.Double pxpdest2 = new Point2D.Double();
      AffineTransform transform = new AffineTransform();
      transform.rotate(theta);
      transform.transform(pxp, pxpdest1);
      
      double pxpY = pxpdest1.getY();
      
      // Rotate result around Y axis
      pxp = new Point2D.Double(pxpdest1.getX(), xp[2]);
      transform = new AffineTransform();
      transform.rotate(phi);
      transform.transform(pxp, pxpdest2);
      
      double pxpX = pxpdest2.getX();
      double pxprime[] = {pxpX, pxpY, 0};
      
      // Dot image of X' with X to get the cos of angle between them  
      double cosrho = MatrixManipulator.dot(e[0], pxprime);
      double rho = Math.acos(cosrho);
      return rho;
    }

  /** 
   * Gets the AffineTransform needed to rotate a BufferedImage slice by 
   * the given angle.  
   * 
   * @param images BufferedImageArray that comprises the volume of image data.
   * @param radians Angle to rotate volume about the axis perpendicular 
   * to the plane represented by an element of the given BufferedImage array.
   * @param spacings Vector to hold the slice spacings.
   * 
   * @return  AffineTransform needed to rotate a BufferedImage slice by 
   * the given angle about the axis perpendicular to the plane represented
   * by an element of the given BufferedImage array.
   */
  private AffineTransform _getAffineTransform(BufferedImage[] images,
					      double radians, 
					      Vector spacings) 
    {
      // Construct the transform
      BufferedImage image = images[0];
      double pivotX = image.getWidth()/2.0;
      double pivotY = image.getHeight()/2.0;
         
      // Determine where to translate after rotating
      int x = image.getMinX();
      int y = image.getMinY();
      int width = image.getWidth();
      int height = image.getHeight();

      // Get the pixel x and y spacing
      double xspacing = Math.abs(image.getTileWidth());
      double yspacing = Math.abs(image.getTileHeight());

      Point tl = new Point(x,y);
      Point tr = new Point(x + width, y);
      Point bl = new Point(x, y + height);
      Point br = new Point(x + width, y + height);
  
      Point2D.Double tldest = new Point2D.Double();
      Point2D.Double trdest = new Point2D.Double();
      Point2D.Double bldest = new Point2D.Double();
      Point2D.Double brdest = new Point2D.Double();
      
      AffineTransform transform = new AffineTransform();

      // 1. Translate the middle of the image to the origin
      transform.preConcatenate(
	     AffineTransform.getTranslateInstance(-1*pivotX, -1*pivotY));
	  
      // 2. Scale Image so that pixel width and height spacing are the same
      if (xspacing > yspacing) {
	AffineTransform scale = new AffineTransform();
	scale.setToScale(xspacing/yspacing, 1.0);
	transform.preConcatenate(scale);
      }
      else if (xspacing < yspacing) {
	AffineTransform scale = new AffineTransform();
	scale.setToScale(1.0, yspacing/xspacing);
	transform.preConcatenate(scale);
      }
       
      // 3. Rotate the image clockwise
      transform.preConcatenate(AffineTransform.getRotateInstance(radians));

      // Determine the image of the 4 corners of the image
      transform.transform(tl, tldest);
      transform.transform(tr, trdest);
      transform.transform(bl, bldest);
      transform.transform(br, brdest);

      // Create the rectangle whose 4 corners are those points
      Polygon p = new Polygon(
	 new int[]{ (int) Math.rint(tldest.x), 
		    (int) Math.rint(trdest.x), 
		    (int) Math.rint(bldest.x), 
		    (int) Math.rint(brdest.x) },
	 new int[]{ (int) Math.rint(tldest.y), 
		    (int) Math.rint(trdest.y), 
		    (int) Math.rint(bldest.y), 
		    (int) Math.rint(brdest.y) },
	 4);   

      // Get the bounding Rectangle for the rotated rectangle
      Rectangle bounds = p.getBounds();

      // Move the top left corner of the bounding Rectangle to the origin
      transform.preConcatenate(
	     AffineTransform.getTranslateInstance(-bounds.x, -bounds.y));

      // Scale down to bounding box of original rectangle rotated by radians
      AffineTransform scale = new AffineTransform();
      double rwidth = width*Math.cos(radians) + height*Math.sin(radians);
      double rheight = width*Math.sin(radians) + height*Math.cos(radians);
      double wAdjustRatio = rwidth/bounds.width;
      double hAdjustRatio = rheight/bounds.height;
      scale.setToScale(wAdjustRatio, hAdjustRatio);
      transform.preConcatenate(scale);

      // Store the spacings
      spacings.clear();
      spacings.add(new Double(wAdjustRatio/Math.min(xspacing, yspacing)));
      spacings.add(new Double(hAdjustRatio/Math.min(xspacing, yspacing)));

      return transform;
  }

  /**
   * Rotates and then vertically flips each of the images.
   *
   * @param images Images to rotate and flip.
   * @param degrees Angle to rotate by.
   * @param spacings Vector to hold the slice spacings.
   *
   * @return Images rotated and then flipped.
   */
  private BufferedImage[] _scaleRotateAndRescale(BufferedImage[] images,
					         double radians, 
						 Vector spacings)
    {
      // Nothing to do
      if (Math.cos(radians) == 1.0 && Math.sin(radians) == 0.0) 
	{ return images; }
      if (images.length == 0) { return images; }

      // Get the AffineTransform
      AffineTransform transform = _getAffineTransform(images, radians,
						      spacings);

      // Construct the transform operator
      AffineTransformOp transformOperator 
	= new AffineTransformOp(transform,
				AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
      // Rotate each image
      BufferedImage[] newImages = new BufferedImage[images.length]; 

      for (int i = 0; i < images.length; i++) {

       	WritableRaster raster = images[i].getRaster();
	 
	// Create a new WriteableRaster to store the transformed image
	// using the appropriate dimensions 
	int width = images[i].getWidth();
	int height = images[i].getHeight();
	double rwidth = width*Math.cos(radians) + height*Math.sin(radians);
	double rheight = width*Math.sin(radians) + height*Math.cos(radians);
	
	int rw = (int) Math.ceil(rwidth);
	int rh = (int) Math.ceil(rheight);
	
	WritableRaster reformattedRaster =
	  raster.createCompatibleWritableRaster(rw,rh);

	// Create new BufferedImage holding the bigger WriteableRaster
	newImages[i] = new BufferedImage(images[0].getColorModel(),
					 reformattedRaster, false,
					 new Hashtable());
	// Transform the image
	newImages[i] = transformOperator.filter(images[i], newImages[i]);
      }
      // Return the new images
      return newImages;
    }

  /**
   * Changes the order of the images in the stack.
   *
   * @param images Images to change the order of.
   *
   * @return Images with their order reversed.
   */
  private BufferedImage[] _reverseOrder(BufferedImage[] images)
    {
      BufferedImage[] newImages = new BufferedImage[images.length];
      for (int i = 0; i < images.length; i++) {
	newImages[i] = images[ images.length - i - 1 ];
      }
      return newImages;
    }

  /** 
   * Returns the data obtained from a writable raster.  
   * 
   * @param wraster WriteableRaster to obtain the data from.
   * 
   * @return Data obtained from the given WritableRaster's DataBuffer or null.
   */
  private Object _getData(WritableRaster wraster) 
    {
      DataBuffer dataBuffer = wraster.getDataBuffer();
      if (dataBuffer instanceof DataBufferByte) {
	return ((DataBufferByte)dataBuffer).getData();
      }
      else if (dataBuffer instanceof DataBufferDouble) {
	return ((DataBufferDouble)dataBuffer).getData();
      }
      else if (dataBuffer instanceof DataBufferFloat) {
	return ((DataBufferFloat)dataBuffer).getData();
      }
      else if (dataBuffer instanceof DataBufferInt) {
	return ((DataBufferInt)dataBuffer).getData();
      }
      else if (dataBuffer instanceof DataBufferShort) {
	return ((DataBufferShort)dataBuffer).getData();
      }
      else if (dataBuffer instanceof DataBufferUShort) {
	return ((DataBufferUShort)dataBuffer).getData();
      }
      return null;
    } 

  /**
   * Reformats the images from the row-column plane to the row-stack plane.
   *
   * @param images Images to reformat.
   *
   * @return Images to reformat from the row-column plane to the row-stack
   *         plane.
   */
  private BufferedImage[] _reformat(BufferedImage[] images)
    {
      // *****
      //
      // Images are reformatted from the row-column to the row-stack plane
      //
      //                             
      //   |                            |
      //   |  ######                    |  ######
      //   |   ######                   |  ######  IMAGE N
      //   |    ######  IMAGE N         |  ######
      //   |                            |  ######      \
      //   *-------        |     ===>   *--------       \
      //    \ ######       |             \     ######
      //     \ ######                     \    ######  IMAGE 0
      //      \ ######  IMAGE 0            \   ######
      //                                       ######
      //
      // *****

      // Nothing to do
      if (images.length == 0) { return images; }

      // Create an array of the Rasters
      WritableRaster[] rasters = new WritableRaster[images.length];
      for (int i = 0; i < images.length; i++) {
	rasters[i] = images[i].getRaster();
      }
      // Reformat the Rasters
      int height = images[0].getHeight();
      BufferedImage[] newImages = new BufferedImage[height];
      for (int i = 0; i < height; i++) {
	WritableRaster reformattedRaster = _getReformattedRaster(rasters, i);
	newImages[i] = new BufferedImage(images[0].getColorModel(),
					 reformattedRaster, false,
					 new Hashtable());
      }
      // Return the reformatted rasters
      return newImages;
    }

  /**
   * Creates a new Writable Raster by combining the pixel data of the specified
   * row of each of the given Writable Rasters.
   *
   * @param rasters Writable Rasters from which to copy pixel data from.
   * @param rowNumber Number of the row of each Raster to copy pixel data from.
   *
   * @return Writable Raster constructed by combining the pixel data of the
   *         specified row of each of the given Writable Rasters.
   */
  private WritableRaster _getReformattedRaster(WritableRaster[] rasters,
					       int rowNumber)
    {
      // Get the width, height, and size of the Rasters
      int width = rasters[0].getWidth();
      int height = rasters[0].getHeight();
      int size = rasters.length;

      // Create the reformatted Raster using the appropriate dimensions
      WritableRaster reformattedRaster =
	rasters[0].createCompatibleWritableRaster(width, size);

      // Get the pixel data buffer of the reformatted Raster
      Object reformattedData = _getData(reformattedRaster);
     
      // Copy the specified row of each Raster to the reformatted Raster
      for (int i = 0; i < size; i++) {
	
	Object data = _getData(rasters[i]);

	// Copy the row
	System.arraycopy(data, rowNumber*width, reformattedData,i*width,width);
      }
      // Return the reformatted Raster
      return reformattedRaster;
    }
}
