package edu.ucla.loni.debabel.processors.keith.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.awt.image.WritableRaster;
import java.util.Hashtable;

/**
 * Processor that reformats a set of Buffered Images in Talairach space.
 *
 * @author Scott C. Neu
 *
 * @version 1 July 2003
 */
public class TalairachImageReformatter extends Processor
{
  /** Allowed Talairach space directions. */
  private static String DIRS = Talairach.POSITIVE_X + ", " + 
                               Talairach.NEGATIVE_X + ", " +
                               Talairach.POSITIVE_Y + ", " +
                               Talairach.NEGATIVE_Y + ", " +
                               Talairach.POSITIVE_Z + ", " +
                               Talairach.NEGATIVE_Z;

  /** Description of this Processor. */
  private static String DESC = "Reformats images in Talairach space.";

  /** Node name for the Talairach coordinate for the image rows. */
  private static String ROW_NAME = "Row Talairach Coordinate";

  /** Node description for the row Talairach coordinate. */
  private static String ROW_DESC = DIRS;

  /** Node name for the Talairach coordinate for the image columns. */
  private static String COL_NAME = "Column Talairach Coordinate";

  /** Node description for the column Talairach coordinate. */
  private static String COL_DESC = DIRS;

  /** Node name for the Talairach coordinate for the image stack. */
  private static String STACK_NAME = "Stack Talairach Coordinate";

  /** Node description for the stack Talairach coordinate. */
  private static String STACK_DESC = DIRS;

  /** Node name for the row step size. */
  private static String ROW_STEP_NAME = "Row Spacing";

  /** Node description for the row step size. */
  private static String ROW_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the rows.";

  /** Node name for the column step size. */
  private static String COL_STEP_NAME = "Column Spacing";

  /** Node description for the column step size. */
  private static String COL_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the columns.";

  /** Node name for the stack step size. */
  private static String STACK_STEP_NAME = "Stack Spacing";

  /** Node description for the stack step size. */
  private static String STACK_STEP_DESC = "Step size between adjacent " +
                                          "voxels in the stack.";

  /** Node name for the images. */
  private static String IMAGES_NAME = "Images";

  /** Node description for the images. */
  private static String IMAGES_DESC = "Images.";

  /**
   * Constructs a TalairachImageReformatter.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TalairachImageReformatter(String name)
    {
      super(name, DESC);

      // Create the Input Nodes for the current Talairach Coordinates
      Class type = String.class;
      InputNode inputNode = new InputNode(ROW_NAME, ROW_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(COL_NAME, COL_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(STACK_NAME, STACK_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Nodes for the new Talairach Coordinates
      type = String.class;
      inputNode = new InputNode("New "+ROW_NAME, ROW_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode("New "+COL_NAME, COL_DESC, type);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode("New "+STACK_NAME, STACK_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Nodes for the spacings
      type = Double.class;
      inputNode = new InputNode(ROW_STEP_NAME, ROW_STEP_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(COL_STEP_NAME, COL_STEP_DESC, type);
      _addInputNode(inputNode);

      type = Double.class;
      inputNode = new InputNode(STACK_STEP_NAME, STACK_STEP_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the images
      type = BufferedImage[].class;
      inputNode = new InputNode(IMAGES_NAME, IMAGES_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Nodes for the spacings
      type = Double.class;
      OutputNode node = new OutputNode("New "+ROW_STEP_NAME,
				       "New "+ROW_STEP_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode("New "+COL_STEP_NAME, "New "+COL_STEP_DESC, type);
      _addOutputNode(node);

      type = Double.class;
      node = new OutputNode("New "+STACK_STEP_NAME, "New "+STACK_STEP_DESC,
			    type);
      _addOutputNode(node);

      // Create the Output Node for the images
      type = BufferedImage[].class;
      node = new OutputNode("New "+IMAGES_NAME, "New "+IMAGES_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Input Node values for the current Talairach Coordinates
	String rowC = null;
	Object value = _getInputValue(ROW_NAME, null);
	if (value instanceof String) { rowC = (String)value; }

	String colC = null;
	value = _getInputValue(COL_NAME, null);
	if (value instanceof String) { colC = (String)value; }

	String stackC = null;
	value = _getInputValue(STACK_NAME, null);
	if (value instanceof String) { stackC = (String)value; }

	// Get the Input Node values for the new Talairach Coordinates
	String newRowC = null;
	value = _getInputValue("New "+ROW_NAME, null);
	if (value instanceof String) { newRowC = (String)value; }

	String newColC = null;
	value = _getInputValue("New "+COL_NAME, null);
	if (value instanceof String) { newColC = (String)value; }

	String newStackC = null;
	value = _getInputValue("New "+STACK_NAME, null);
	if (value instanceof String) { newStackC = (String)value; }

	// Get the Input Node values for the spacings
	Double rowSpacing = null;
	value = _getInputValue(ROW_STEP_NAME, null);
	if (value instanceof Double) { rowSpacing = (Double)value; }

	Double colSpacing = null;
	value = _getInputValue(COL_STEP_NAME, null);
	if (value instanceof Double) { colSpacing = (Double)value; }

	Double stackSpacing = null;
	value = _getInputValue(STACK_STEP_NAME, null);
	if (value instanceof Double) { stackSpacing = (Double)value; }

	// Get the Input Node value for the images
	BufferedImage[] images = null;
	value = _getInputValue(IMAGES_NAME, null);
	if (value instanceof BufferedImage[]) {images=(BufferedImage[])value;}

	// Determine if enough information is provided
	BufferedImage[] newImages = null;
	if ( Talairach.isValid(rowC) && Talairach.isValid(colC) &&
	     Talairach.isValid(stackC) && Talairach.isValid(newRowC) &&
	     Talairach.isValid(newColC) && Talairach.isValid(newStackC) &&
	     images != null)
	  {
	    newImages = images;
	    String newRowVar = Talairach.getVariable(newRowC);
	    String newColVar = Talairach.getVariable(newColC);
	    String newStackVar = Talairach.getVariable(newStackC);

	    // Current stack variable is new row or column variable
	    //
	    //    Vrow               Vcolumn
	    //    |                  |  
	    //    *---     OR        *---  
	    //   /                  /    
	    //
	    if ( newRowVar.equals( Talairach.getVariable(stackC) ) ||
		 newColVar.equals( Talairach.getVariable(stackC) ) )
	      {

	      // Current row variable is new stack variable
	      //    
	      //    |  
	      //    *--- Vstack
	      //   /    
	      //  
	      if ( newStackVar.equals( Talairach.getVariable(rowC) ) ) {

		// Rotate the row direction to the column direction
		newImages = _rotateAndFlip(newImages, 1, false);

		// Exchange the row and column directions
		String temp = colC;
		colC = rowC;
		rowC = temp;

		// Reverse the sign of the new row direction
		rowC = Talairach.reverseSign(rowC);

		// Exchange the row and column spacings
		Double tempSpacing = colSpacing;
		colSpacing = rowSpacing;
		rowSpacing = tempSpacing;
	      }

	      // Current column variable is new stack variable
	      //
	      //    |  
	      //    *---
	      //   /    
	      //  Vstack

	      // Reformat from the row-column plane to the row-stack plane
	      newImages = _reformat(newImages);

	      // Exchange the stack and column directions
	      String temp = stackC;
	      stackC = colC;
	      colC = temp;

	      // Reverse the sign of the new stack direction
	      stackC = Talairach.reverseSign(stackC);

	      // Exchange the stack and column spacings
	      Double tempSpacing = stackSpacing;
	      stackSpacing = colSpacing;
	      colSpacing = tempSpacing;
	    }

	    // Current stack variable is new stack variable
	    //
	    //    Vstack
	    //    |  
	    //    *---
	    //   /    

	    // Get the number of rotations to align the column direction
	    int numberOfRots = _getRotations(rowC, colC, newColC);

	    // Correct the row and column directions
	    if (numberOfRots % 2 == 1) {

	      // Exchange the row and column directions
	      String temp = colC;
	      colC = rowC;
	      rowC = temp;

	      // Exchange the row and column spacings
	      Double tempSpacing = colSpacing;
	      colSpacing = rowSpacing;
	      rowSpacing = tempSpacing;
	    }

	    // Correct the row and column signs
	    if (numberOfRots == 1) { rowC = Talairach.reverseSign(rowC); }
	    else if (numberOfRots == 2) {
	      rowC = Talairach.reverseSign(rowC);
	      colC = Talairach.reverseSign(colC);
	    }
	    else if (numberOfRots == 3) { colC = Talairach.reverseSign(colC); }

	    // Current column variable is new column variable
	    //
	    //    Vstack
	    //    |  
	    //    *---
	    //   /    
	    //  Vcolumn

	    // Determine if a vertical flip is necessary
	    boolean needsVflip = false;
	    if ( !newRowC.equals(rowC) ) { needsVflip = true; }

	    // Rotate and flip each image in the stack
	    newImages = _rotateAndFlip(newImages, numberOfRots, needsVflip);

	    // Change the stack order if required
	    if ( !newStackC.equals(stackC) ) {
	      newImages = _reverseOrder(newImages);
	    }
	  }

	// Set the output values
	_setOutputValue("New "+ROW_STEP_NAME, rowSpacing);
	_setOutputValue("New "+COL_STEP_NAME, colSpacing);
	_setOutputValue("New "+STACK_STEP_NAME, stackSpacing);
	_setOutputValue("New "+IMAGES_NAME, newImages);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Constructs a TalairachImageReformatter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public TalairachImageReformatter(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Rotates and then vertically flips each of the images.
   *
   * @param images Images to rotate and flip.
   * @param numberOfRotations Number of 90 degree clockwise rotations.
   * @param needsVflip True if each image is to be vertically flipped after
   *                   rotating.
   *
   * @return Images rotated and then flipped.
   */
  private BufferedImage[] _rotateAndFlip(BufferedImage[] images,
					 int numberOfRotations,
					 boolean needsVflip)
    {
      // Nothing to do
      if (numberOfRotations == 0 && needsVflip == false) { return images; }
      if (images.length == 0) { return images; }

      // Construct the transform
      BufferedImage image = images[0];
      double pivotX = image.getWidth()/2.0;
      double pivotY = image.getHeight()/2.0;
      AffineTransform transform = new AffineTransform();

      // Translate the image back
      if (numberOfRotations % 2 == 0) { transform.translate(pivotX, pivotY); }
      else { transform.translate(pivotY, pivotX); }
      
      // Flip the image about the Y axis
      if (needsVflip) {
	AffineTransform vFlip = new AffineTransform();
	vFlip.setToScale(-1.0, 1.0);
	transform.concatenate(vFlip);
      }

      // Rotate the image clockwise
      double theta = Math.toRadians(90*numberOfRotations);
      transform.rotate(theta);
	  
      // Translate the middle of the image to the origin
      transform.translate(-1*pivotX, -1*pivotY);

      // Construct the transform operator
      AffineTransformOp transformOperator 
	= new AffineTransformOp(transform,
				AffineTransformOp.TYPE_NEAREST_NEIGHBOR);

      // Rotate and flip each image
      BufferedImage[] newImages = new BufferedImage[images.length];
      for (int i = 0; i < images.length; i++) {
	newImages[i] = transformOperator.filter(images[i], null);
      }

      // Return the new images
      return newImages;
    }

  /**
   * Changes the order of the images in the stack.
   *
   * @param images Images to change the order of.
   *
   * @return Images with their order reversed.
   */
  private BufferedImage[] _reverseOrder(BufferedImage[] images)
    {
      BufferedImage[] newImages = new BufferedImage[images.length];
      for (int i = 0; i < images.length; i++) {
	newImages[i] = images[ images.length - i - 1 ];
      }

      return newImages;
    }

  /**
   * Reformats the images from the row-column plane to the row-stack plane.
   *
   * @param images Images to reformat.
   *
   * @return Images to reformat from the row-column plane to the row-stack
   *         plane.
   */
  private BufferedImage[] _reformat(BufferedImage[] images)
    {
      // *****
      //
      // Images are reformatted from the row-column to the row-stack plane
      //
      //                             
      //   |                            |
      //   |  ######                    |  ######
      //   |   ######                   |  ######  IMAGE N
      //   |    ######  IMAGE N         |  ######
      //   |                            |  ######      \
      //   *-------        |     ===>   *--------       \
      //    \ ######       |             \     ######
      //     \ ######                     \    ######  IMAGE 0
      //      \ ######  IMAGE 0            \   ######
      //                                       ######
      //
      // *****

      // Nothing to do
      if (images.length == 0) { return images; }

      // Create an array of the Rasters
      WritableRaster[] rasters = new WritableRaster[images.length];
      for (int i = 0; i < images.length; i++) {
	rasters[i] = images[i].getRaster();
      }

      // Reformat the Rasters
      int height = images[0].getHeight();
      BufferedImage[] newImages = new BufferedImage[height];
      for (int i = 0; i < height; i++) {
	WritableRaster reformattedRaster = _getReformattedRaster(rasters, i);
	newImages[i] = new BufferedImage(images[0].getColorModel(),
					 reformattedRaster, false,
					 new Hashtable());
      }

      // Reverse the order of the images
      newImages = _reverseOrder(newImages);

      // Return the reformatted rasters
      return newImages;
    }

  /**
   * Creates a new Writable Raster by combining the pixel data of the specified
   * row of each of the given Writable Rasters.
   *
   * @param rasters Writable Rasters from which to copy pixel data from.
   * @param rowNumber Number of the row of each Raster to copy pixel data from.
   *
   * @return Writable Raster constructed by combining the pixel data of the
   *         specified row of each of the given Writable Rasters.
   */
  private WritableRaster _getReformattedRaster(WritableRaster[] rasters,
					       int rowNumber)
    {
      // Get the width, height, and size of the Rasters
      int width = rasters[0].getWidth();
      int height = rasters[0].getHeight();
      int size = rasters.length;

      // Create the reformatted Raster using the appropriate dimensions
      WritableRaster reformattedRaster =
	rasters[0].createCompatibleWritableRaster(width, size);

      // Get the pixel data buffer of the reformatted Raster
      Object reformattedData = null;
      DataBuffer reformattedDataBuffer = reformattedRaster.getDataBuffer();
      if (reformattedDataBuffer instanceof DataBufferByte) {
	reformattedData = ((DataBufferByte)reformattedDataBuffer).getData();
      }
      else if (reformattedDataBuffer instanceof DataBufferDouble) {
	reformattedData = ((DataBufferDouble)reformattedDataBuffer).getData();
      }
      else if (reformattedDataBuffer instanceof DataBufferFloat) {
	reformattedData = ((DataBufferFloat)reformattedDataBuffer).getData();
      }
      else if (reformattedDataBuffer instanceof DataBufferInt) {
	reformattedData = ((DataBufferInt)reformattedDataBuffer).getData();
      }
      else if (reformattedDataBuffer instanceof DataBufferShort) {
	reformattedData = ((DataBufferShort)reformattedDataBuffer).getData();
      }
      else if (reformattedDataBuffer instanceof DataBufferUShort) {
	reformattedData = ((DataBufferUShort)reformattedDataBuffer).getData();
      }

      // Copy the specified row of each Raster to the reformatted Raster
      for (int i = 0; i < size; i++) {
	Object data = null;
	DataBuffer dataBuffer = rasters[i].getDataBuffer();
	if (dataBuffer instanceof DataBufferByte) {
	  data = ((DataBufferByte)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferDouble) {
	  data = ((DataBufferDouble)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferFloat) {
	  data = ((DataBufferFloat)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferInt) {
	  data = ((DataBufferInt)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferShort) {
	  data = ((DataBufferShort)dataBuffer).getData();
	}
	else if (dataBuffer instanceof DataBufferUShort) {
	  data = ((DataBufferUShort)dataBuffer).getData();
	}

	// Copy the row
	System.arraycopy(data, rowNumber*width, reformattedData, i*width,
			 width);
      }

      // Return the reformatted Raster
      return reformattedRaster;
    }

  /**
   * Gets the number of rotations needed to align the column coordinates.
   *
   * @param rowC1 First row coordinate.
   * @param colC1 First column coordinate.
   * @param colC2 Second column coordinate.
   *
   * @return Number of 90 degree clockwise rotations to apply to the first
   *         coordinates to align the first's column coordinate with the
   *         second.
   */
  private int _getRotations(String rowC1, String colC1, String colC2)
    {
      String colC2Var = Talairach.getVariable(colC2);

      // Columns coordinates are parallel
      if ( colC2Var.equals( Talairach.getVariable(colC1) ) ) {

	// If they have the same sign, they are aligned
	if ( Talairach.isPositive(colC2) == Talairach.isPositive(colC1) ) {
	  return 0;
	}

	// Otherwise they point in opposite directions
	else { return 2; }
      }

      // Columns coordinates are perpendicular
      else {

	// 1st row coordinate has the same sign as the 2nd column coordinate
	if ( Talairach.isPositive(colC2) == Talairach.isPositive(rowC1) ) {
	  return 1;
	}

	// Otherwise more rotations are necessary
	else { return 3; }
      }
    }
}
