/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that produces file names for the Analyze file format.
 *
 * @version 10 May 2003
 */
public class AnalyzeFileNamer extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Produces file names for the ANALYZE file " +
                               "format.";

  /** Name of the Input Node that accepts the base file name. */
  private static String BASE_NODE_NAME = "Base File Name";

  /** Description of the Input Node that accepts the base file name. */
  private static String BASE_NODE_DESC = "File name from which to construct " +
                                         "the ANALYZE file names.";

  /** Name of the Output Node that returns the header file name. */
  private static String HDR_NODE_NAME = "Header File Name";

  /** Description of the Output Node that returns the header file name. */
  private static String HDR_NODE_DESC = "Header file name for the ANALYZE " +
                                        "file format.";

  /** Name of the Output Node that returns the image file name. */
  private static String IMG_NODE_NAME = "Image File Name";

  /** Description of the Output Node that returns the image file name. */
  private static String IMG_NODE_DESC = "Image file name for the ANALYZE " +
                                        "file format.";

  /**
   * Constructs an Analyze File Namer with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public AnalyzeFileNamer(String name)
    {
      super(name, DESC);

      // Create the Input Node that accepts the base file name
      Class type = String.class;
      InputNode baseNode = new InputNode(BASE_NODE_NAME, BASE_NODE_DESC, type);
      _addInputNode(baseNode);

      // Create the Output Node that returns the header file name
      type = String.class;
      OutputNode hdrNode = new OutputNode(HDR_NODE_NAME, HDR_NODE_DESC, type);
      _addOutputNode(hdrNode);

      // Create the Output Node that returns the image file name
      type = String.class;
      OutputNode imgNode = new OutputNode(IMG_NODE_NAME, IMG_NODE_DESC, type);
      _addOutputNode(imgNode);
    }

  /**
   * Constructs an Analyze File Namer using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public AnalyzeFileNamer(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the base file name
	String baseFileName = (String)_getInputValue(BASE_NODE_NAME, null);
	if (baseFileName != null) {

	  // Set the header and image file names by adding "hdr" and "img"
	  _setOutputValue(HDR_NODE_NAME, baseFileName + ".hdr");
	  _setOutputValue(IMG_NODE_NAME, baseFileName + ".img");
	}
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
