/*
COPYRIGHT NOTICE
Copyright (c) 2004  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import edu.ucla.loni.imageio.Utilities;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.util.Vector;
import javax.imageio.ImageTypeSpecifier;

/**
 * Processor that converts a Buffered Image of one type into a Buffered Image
 * of another type.
 *
 * @version 22 January 2004
 */
public class BufferedImageConverter extends Processor
{
  /** Description of this Processor. */
  private static String DESC =
  "Converts a Buffered Image of one data type into a different data type.";

  /** Input Node Name for the image to convert. */
  private static String ORIG_IMAGE_NAME = "Original Image";

  /** Input Node Description for the image to convert. */
  private static String ORIG_IMAGE_DESC =
  "Image to convert to another data type.";

  /** Output Node Name for the converted image. */
  private static String NEW_IMAGE_NAME = "Converted Image";

  /** Output Node Description for the converted primitive. */
  private static String NEW_IMAGE_DESC = "Image after conversion.";

  /** Property for the data type. */
  private static String PROP_NAME = "type";

  /** Property description for the data type. */
  private static String PROP_DESC = "Data type to convert to.";

  /**
   * Constructs a Buffered Image Converter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public BufferedImageConverter(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Buffered Image Converter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public BufferedImageConverter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node
      Class type = BufferedImage.class;
      InputNode inputNode = new InputNode(ORIG_IMAGE_NAME, ORIG_IMAGE_DESC,
					  type);
      _addInputNode(inputNode);

      // Create the Output Node
      type = BufferedImage.class;
      OutputNode node = new OutputNode(NEW_IMAGE_NAME, NEW_IMAGE_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the original Buffered Image
	Object value = _getInputValue(ORIG_IMAGE_NAME, null);

	// Convert the Buffered Image to the specified data type
	if (value instanceof BufferedImage) {
	  BufferedImage origImage = (BufferedImage)value;

	  // Get the Data Buffer
	  DataBuffer dataBuffer = origImage.getRaster().getDataBuffer();

	  // Already of short data type; nothing to do
	  if (dataBuffer instanceof DataBufferShort ||
	      dataBuffer instanceof DataBufferUShort) {
	    _setOutputValue(NEW_IMAGE_NAME, origImage);
	  }

	  // Convert bytes to shorts
	  else if (dataBuffer instanceof DataBufferByte) {

	    // Get the bytes
	    byte[] byteData = ((DataBufferByte)dataBuffer).getData();

	    // Copy the byte array to a short array
	    short[] shortData = new short[ byteData.length ];
	    for (int i = 0; i < byteData.length; i++) {
	      shortData[i] = (short)(byteData[i] & 0x00ff);
	    }

	    // Create a new Raster for the short array
	    int w = origImage.getWidth();
	    int h = origImage.getHeight();
	    ImageTypeSpecifier spec = Utilities.getGrayImageType(w, h, 16);
	    WritableRaster newRaster =
	      Raster.createWritableRaster( spec.getSampleModel(w, h),
					   new Point(0, 0) );

	    // Create a new Buffered Image
	    BufferedImage newImage = Utilities.getGrayUshortImage(shortData,
								  newRaster);
	    _setOutputValue(NEW_IMAGE_NAME, newImage);
	  }
	}
      }

      // Unable to convert
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Data Types
      Vector optionSet = new Vector();
      optionSet.add( Short.class );

      // Create a Set Restriction
      String name = "Allowed Data Types";
      String desc = "Data types that can be converted to.";
      Restriction restriction = new SetRestriction(name, desc, optionSet);

      // Default primitive type
      name = PROP_NAME;
      desc = PROP_DESC;
      Object value = Short.class;
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
