/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.awt.image.WritableRaster;
import java.util.Hashtable;

/**
 * Processor that creates a smaller Buffered Image from a Buffered Image.
 *
 * @version 15 September 2003
 */
public class BufferedImageDivider extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Makes a smaller image from a Buffered Image.";

  /** Input Node name for the Buffered Image. */
  private static String IMAGE_NAME = "Original Image";

  /** Input Node description for the Buffered Image. */
  private static String IMAGE_DESC = "Image to make a smaller image from.";

  /** Input Node name for the upper-left X coordinate. */
  private static String X_NAME = "X";

  /** Input Node description for the upper-left X coordinate. */
  private static String X_DESC = "Upper-left X coordinate specifying the " +
                                 "rectangle of the smaller image.";

  /** Input Node name for the upper-left Y coordinate. */
  private static String Y_NAME = "Y";

  /** Input Node description for the upper-left Y coordinate. */
  private static String Y_DESC = "Upper-left Y coordinate specifying the " +
                                 "rectangle of the smaller image.";

  /** Input Node name for the width. */
  private static String WIDTH_NAME = "Width";

  /** Input Node description for the width. */
  private static String WIDTH_DESC = "Width of the smaller image.";

  /** Input Node name for the height. */
  private static String HEIGHT_NAME = "Height";

  /** Input Node description for the height. */
  private static String HEIGHT_DESC = "Height of the smaller image.";

  /** Output Node name for the smaller Buffered Image. */
  private static String SMALL_IMAGE_NAME = "Smaller Image";

  /** Output Node description for the smaller Buffered Image. */
  private static String SMALL_IMAGE_DESC = "Smaller image created from the " +
                                           "original image.";

  /**
   * Constructs a Buffered Image Divider with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public BufferedImageDivider(String name)
    {
      super(name, DESC);

      // Create the positive Restriction
      String rName = "Positive Integers";
      String desc = "Only integers greater than zero are allowed.";
      RangeRestriction r = new RangeRestriction(rName, desc, 1,
						Double.POSITIVE_INFINITY);

      // Create the Input Node for the Buffered Image
      Class type = BufferedImage.class;
      InputNode inputNode = new InputNode(IMAGE_NAME, IMAGE_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the upper-left X coordinate
      type = Integer.class;
      inputNode = new InputNode(X_NAME, X_DESC, type);
      inputNode.setRestriction( Utilities.getPositiveRestriction() );
      _addInputNode(inputNode);

      // Create the Input Node for the upper-left Y coordinate
      type = Integer.class;
      inputNode = new InputNode(Y_NAME, Y_DESC, type);
      inputNode.setRestriction( Utilities.getPositiveRestriction() );
      _addInputNode(inputNode);

      // Create the Input Node for the width
      type = Integer.class;
      inputNode = new InputNode(WIDTH_NAME, WIDTH_DESC, type);
      inputNode.setRestriction(r);
      _addInputNode(inputNode);

      // Create the Input Node for the height
      type = Integer.class;
      inputNode = new InputNode(HEIGHT_NAME, HEIGHT_DESC, type);
      inputNode.setRestriction(r);
      _addInputNode(inputNode);

      // Create the Output Node for the smaller Buffered Image
      type = BufferedImage.class;
      OutputNode node = new OutputNode(SMALL_IMAGE_NAME,SMALL_IMAGE_DESC,type);
      _addOutputNode(node);
    }

  /**
   * Constructs a Buffered Image Divider using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public BufferedImageDivider(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Input value
	BufferedImage image = (BufferedImage)_getInputValue(IMAGE_NAME, null);
	if (image == null) { return; }

	// Get the dividing rectangle coordinates
	Integer x = (Integer)_getInputValue(X_NAME, null);
	Integer y = (Integer)_getInputValue(Y_NAME, null);
	Integer width = (Integer)_getInputValue(WIDTH_NAME, null);
	Integer height = (Integer)_getInputValue(HEIGHT_NAME, null);

	if (x == null || y == null || width == null || height == null) {
	  return;
	}

	// Get the dividing rectangle
	Rectangle rect = new Rectangle( x.intValue(), y.intValue(),
					width.intValue(), height.intValue() );

	// Check the validity of the dividing rectangle
	Rectangle r = new Rectangle( image.getWidth(), image.getHeight() );
	if ( !r.contains(rect) ) {
	  String msg = "Smaller image does not lie completely within the " +
	               "original image.";
	  throw new IllegalArgumentException(msg);
	}

	// Make a copy of the smaller image's data (move origin to 0,0)
	WritableRaster raster = (WritableRaster)image.getData(rect);
	raster = raster.createWritableTranslatedChild(0, 0);

	// Make a new Buffered Image from the smaller image's data
	BufferedImage smallImage =
	  new BufferedImage( image.getColorModel(), raster,
			     image.getColorModel().isAlphaPremultiplied(),
			     new Hashtable() );

	// Set the new Buffered Image
	_setOutputValue(SMALL_IMAGE_NAME, smallImage);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
