/*
  COPYRIGHT NOTICE
  Copyright (c) 2009  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;

/**
 * Processor that returns a histogram of pixel intensities as well as properties
 * about the pixel data characteristics.
 *
 * @version 2 March 2009
 */
public class BufferedImageHistogram extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Gets histogram properties of image data.";

    /** Input Node name for the images to analyze. */
    private static String IMAGES_NAME = "Images";

    /** Input Node description for the images to analyze. */
    private static String IMAGES_DESC = "Images to analyze.";

    /** Input Node name for the number of histogram bins. */
    private static String BIN_NAME = "Number of Bins";

    /** Input Node description for the number of histogram bins. */
    private static String BIN_DESC = "Number of histogram bins to create.";

    /** Output Node name for the minimum pixel value. */
    private static String MIN_NAME = "Minimum";

    /** Output Node description for the minimum pixel value. */
    private static String MIN_DESC = "Minimum pixel value.";

    /** Output Node name for the maximum pixel value. */
    private static String MAX_NAME = "Maximum";

    /** Output Node description for the maximum pixel value. */
    private static String MAX_DESC = "Maximum pixel value.";

    /** Output Node name for the mean pixel value. */
    private static String MEAN_NAME = "Mean";

    /** Output Node description for the mean pixel value. */
    private static String MEAN_DESC = "Mean pixel value.";

    /** Output Node name for the variance of the pixel values. */
    private static String VAR_NAME = "Variance";

    /** Output Node description for the variance of the pixel values. */
    private static String VAR_DESC = "Variance of the pixel values.";

    /** Output Node name for the standard deviation of the pixel values. */
    private static String STD_NAME = "Standard Deviation";

    /** Output Node desc for the standard deviation of the pixel values. */
    private static String STD_DESC = "Standard deviation of the pixel values.";

    /** Output Node name for the width of a histogram bin. */
    private static String BIN_WIDTH_NAME = "Bin Width";

    /** Output Node description for the width of a histogram bin. */
    private static String BIN_WIDTH_DESC = "Width of each histogram bin.";

    /** Output Node name for the number of histogram bin counts. */
    private static String CNTS_NAME = "Number of Bin Counts";

    /** Output Node description for the number of histogram bin counts. */
    private static String CNTS_DESC = "Number of counts in each histogram bin.";

    /**
     * Constructs a Buffered Image Histogram.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public BufferedImageHistogram(String name)
    {
	super(name, DESC);

      // Create the Input Node for the Buffered Images
      Class type = BufferedImage[].class;
      InputNode inputNode = new InputNode(IMAGES_NAME, IMAGES_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node for the number of histogram bins
      type = Integer.class;
      inputNode = new InputNode(BIN_NAME, BIN_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the minimum pixel value
      type = Double.class;
      OutputNode node = new OutputNode(MIN_NAME, MIN_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the maximum pixel value
      type = Double.class;
      node = new OutputNode(MAX_NAME, MAX_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the mean pixel value
      type = Double.class;
      node = new OutputNode(MEAN_NAME, MEAN_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the variance of the pixel values
      type = Double.class;
      node = new OutputNode(VAR_NAME, VAR_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the standard deviation of the pixel values
      type = Double.class;
      node = new OutputNode(STD_NAME, STD_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the histogram bin width
      type = Double.class;
      node = new OutputNode(BIN_WIDTH_NAME, BIN_WIDTH_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the number of histogram bin counts
      type = Long[].class;
      node = new OutputNode(CNTS_NAME, CNTS_DESC, type);
      _addOutputNode(node);
    }

    /**
     * Constructs a Buffered Image Histogram using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public BufferedImageHistogram(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the images
	    BufferedImage[] images=(BufferedImage[])_getInputValue(IMAGES_NAME,
								   null);
	    if (images == null || images.length == 0) { return; }

	    // Get the number of histogram bins
	    Integer numOfBins = (Integer)_getInputValue(BIN_NAME, null);
	    int numberOfBins = 256;
	    if (numOfBins != null) { numberOfBins = numOfBins.intValue(); }
	    if (numberOfBins == 0) { return; }

	    // Determine basic statistics
	    long count = 0;
	    double min = Double.POSITIVE_INFINITY;
	    double max = Double.NEGATIVE_INFINITY;
	    double mean = 0;
	    for (int b = 0; b < images.length; b++) {
		DataBuffer dataBuffer = images[b].getRaster().getDataBuffer();
		for (int i = 0; i < dataBuffer.getSize(); i++) {
		    double value = dataBuffer.getElemDouble(i);

		    if (value < min) { min = value; }
		    if (value > max) { max = value; }
		    mean += value;
		    count++;
		}
	    }
	    if (count > 0) { mean /= count; }

	    // Initialize the bins
	    long[] bins = new long[numberOfBins];
	    double binWidth = (max-min)/numberOfBins;

	    // Change the bin size if the resolution is poor
	    if (_isIntegral(images[0]) && binWidth < 1) {
		binWidth = 1;
	    }

	    // Fill the bins and compute the variance
	    double variance = 0;
	    for (int b = 0; b < images.length; b++) {
		DataBuffer dataBuffer = images[b].getRaster().getDataBuffer();
		for (int i = 0; i < dataBuffer.getSize(); i++) {
		    double value = dataBuffer.getElemDouble(i);

		    variance += (value-mean)*(value-mean);

		    // Find the bin for the value
		    int binIndex = (int)Math.floor((value-min)/binWidth);
		    if (binIndex < 0) { bins[0]++; }
		    else if (binIndex >= bins.length) { bins[bins.length-1]++; }
		    else { bins[binIndex]++; }
		}
	    }
	    if (count > 1) { variance /= (count-1); }

	    // Set the output values
	    _setOutputValue(MIN_NAME, new Double(min));
	    _setOutputValue(MAX_NAME, new Double(max));
	    _setOutputValue(MEAN_NAME, new Double(mean));
	    _setOutputValue(VAR_NAME, new Double(variance));
	    _setOutputValue(STD_NAME, new Double(Math.sqrt(variance)));
	    _setOutputValue(BIN_WIDTH_NAME, new Double(binWidth));

	    Long[] cnts = new Long[bins.length];
	    for (int i = 0; i < cnts.length; i++) {
		cnts[i] = new Long(bins[i]);
	    }
	    _setOutputValue(CNTS_NAME, cnts);
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }

    /**
     * Determines if the image data is integral.
     *
     * @param image Image.
     *
     * @return True if the image data is integral; false otherwise.
     */
    private boolean _isIntegral(BufferedImage image)
    {
	DataBuffer dataBuffer = image.getRaster().getDataBuffer();

	if (dataBuffer instanceof DataBufferByte) { return true; }
	else if (dataBuffer instanceof DataBufferShort) { return true; }
	else if (dataBuffer instanceof DataBufferUShort) { return true; }
	else if (dataBuffer instanceof DataBufferInt) { return true; }
	return false;

    }
}
