/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;

/**
 * Processor that returns properties of a Buffered Image.
 *
 * @version 12 May 2003
 */
public class BufferedImageInspector extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Extracts properties of a Buffered Image.";

  /** Input Node name for the Buffered Image. */
  private static String IMAGE_NAME = "Image";

  /** Input Node description for the Buffered Image. */
  private static String IMAGE_DESC = "Image to get properties of.";

  /** Output Node name for the Buffered Image width. */
  private static String WIDTH_NAME = "Width";

  /** Output Node description for the Buffered Image width. */
  private static String WIDTH_DESC = "Width of the image.";

  /** Output Node name for the Buffered Image height. */
  private static String HEIGHT_NAME = "Height";

  /** Output Node description for the Buffered Image height. */
  private static String HEIGHT_DESC = "Height of the image.";

  /**
   * Constructs a Buffered Image Inspector with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public BufferedImageInspector(String name)
    {
      super(name, DESC);

      // Create the Input Node for the Buffered Image
      Class type = BufferedImage.class;
      InputNode inputNode = new InputNode(IMAGE_NAME, IMAGE_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the Buffered Image width
      type = Integer.class;
      OutputNode node = new OutputNode(WIDTH_NAME, WIDTH_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the Buffered Image height
      type = Integer.class;
      node = new OutputNode(HEIGHT_NAME, HEIGHT_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Constructs a Buffered Image Inspector using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public BufferedImageInspector(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Input value
	BufferedImage image = (BufferedImage)_getInputValue(IMAGE_NAME, null);

	// Get properties only if an image is present
	Integer width = null;
	Integer height = null;
	if (image != null) {
	  width = new Integer( image.getWidth() );
	  height = new Integer( image.getHeight() );
	}

	// Set the properties
	_setOutputValue(WIDTH_NAME, width);
	_setOutputValue(HEIGHT_NAME, height);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
