/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;

/**
 * Processor that returns the maximum and minimum values of a Buffered Image.
 *
 * @version 20 June 2003
 */
public class BufferedImagePixelRange extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Gets the minimum and maximum pixel value.";

  /** Input Node name for the Buffered Image. */
  private static String IMAGE_NAME = "Image";

  /** Input Node description for the Buffered Image. */
  private static String IMAGE_DESC = "Image to get properties of.";

  /** Output Node name for the minimum pixel value. */
  private static String MIN_NAME = "Minimum";

  /** Output Node description for the minimum pixel value. */
  private static String MIN_DESC = "Minimum pixel value of the image.";

  /** Output Node name for the maximum pixel value. */
  private static String MAX_NAME = "Maximum";

  /** Output Node description for the maximum pixel value. */
  private static String MAX_DESC = "Maximum pixel value of the image.";

  /**
   * Constructs a Buffered Image Pixel Range with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public BufferedImagePixelRange(String name)
    {
      super(name, DESC);

      // Create the Input Node for the Buffered Image
      Class type = BufferedImage.class;
      InputNode inputNode = new InputNode(IMAGE_NAME, IMAGE_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the minimum pixel value
      type = Double.class;
      OutputNode node = new OutputNode(MIN_NAME, MIN_DESC, type);
      _addOutputNode(node);

      // Create the Output Node for the maximum pixel value
      type = Double.class;
      node = new OutputNode(MAX_NAME, MAX_DESC, type);
      _addOutputNode(node);
    }

  /**
   * Constructs a Buffered Image Pixel Range using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public BufferedImagePixelRange(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the Input value
	BufferedImage image = (BufferedImage)_getInputValue(IMAGE_NAME, null);

	// Compute the minimum and maximum pixel value
	Double minimum = null;
	Double maximum = null;

	if (image != null) {
	  double min = Double.POSITIVE_INFINITY;
	  double max = Double.NEGATIVE_INFINITY;
	  DataBuffer dataBuffer = image.getRaster().getDataBuffer();

	  // Find the minimum and maximum
	  for (int i = 0; i < dataBuffer.getSize(); i++) {
	    double pixelValue = dataBuffer.getElemDouble(i);
	    if (pixelValue < min) { min = pixelValue; }
	    if (pixelValue > max) { max = pixelValue; }
	  }

	  // Reset the minimum and maximum
	  minimum = new Double(min);
	  maximum = new Double(max);
	}

	// Set the Output values
	_setOutputValue(MIN_NAME, minimum);
	_setOutputValue(MAX_NAME, maximum);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
