/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.loop.Stateful;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Stateful Processor that maintains the maximum or minimum values inside a
 * Loop Module.
 *
 * @version 23 June 2003
 */
public class ExtremumFinder extends Processor implements Stateful
{
  /** Description of this Processor. */
  private static String DESC = "Determines the minimum and maximum values " +
                               "inside a loop.";

  /** Input Node name for the value to check. */
  private static String VALUE_NAME = "Value";

  /** Input Node description for the value to check. */
  private static String VALUE_DESC = "Value to check against the current " +
                                     "extremum.";

  /** Output Node name for the minimum value. */
  private static String MINIMUM_NAME = "Mininum";

  /** Output Node description for the minimum value. */
  private static String MINIMUM_DESC = "Minimum found from the input values.";

  /** Output Node name for the maximum value. */
  private static String MAXIMUM_NAME = "Maximum";

  /** Output Node description for the maximum value. */
  private static String MAXIMUM_DESC = "Maximum found from the input values.";

  /** Mininum value. */
  private double _min;

  /** Maximum value. */
  private double _max;

  /**
   * Constructs an Extremum Finder with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public ExtremumFinder(String name)
    {
      super(name, DESC);

      // Create the Input Node for the input value
      Class type = Double.class;
      InputNode inputNode = new InputNode(VALUE_NAME, VALUE_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the minimum value
      type = Double.class;
      OutputNode outputNode = new OutputNode(MINIMUM_NAME, MINIMUM_DESC, type);
      _addOutputNode(outputNode);

      // Create the Output Node for the maximum value
      type = Double.class;
      outputNode = new OutputNode(MAXIMUM_NAME, MAXIMUM_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Constructs an Extremum Finder using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public ExtremumFinder(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Reinitializes the state.  This is called by a Loop Module before a loop is
   * started.
   */
  public void reinitializeState()
    {
      _min = Double.POSITIVE_INFINITY;
      _max = Double.NEGATIVE_INFINITY;
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the input value
	Double value = (Double)_getInputValue(VALUE_NAME, null);

	// Update the minimum and maximum
	if (value != null) {
	  double v = value.doubleValue();

	  if (v < _min) { _min = v; }
	  if (v > _max) { _max = v; }
	}

	// Set the Output Nodes
	_setOutputValue(MINIMUM_NAME, new Double(_min));
	_setOutputValue(MAXIMUM_NAME, new Double(_max));
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
