/*
  COPYRIGHT NOTICE
  Copyright (c) 2010  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferShort;
import java.awt.image.DataBufferUShort;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;
import javax.imageio.ImageReadParam;
import javax.imageio.ImageReader;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.stream.FileImageInputStream;
import javax.imageio.stream.FileImageOutputStream;

/**
 * Image Reader that reads images from a temporary file.
 *
 * @version 17 November 2010
 */
public class FileCacheImageReader extends ImageReader
{
    /** Temporary file. */
    private File _tempFile;

    /** Width of an image. */
    private int _imageWidth;

    /** Height of an image. */
    private int _imageHeight;

    /** Image color models. */
    private ColorModel[] _colorModels;

    /** Image sample models. */
    private SampleModel[] _sampleModels;

    /**
     * Constructs a File Cache Image Reader.
     *
     * @param dir Directory to write the temporary file.
     * @param images Images to write to the temporary file.
     *
     * @throws IOException If an I/O error occurs.
     */
    public FileCacheImageReader(File dir, BufferedImage[] images)
	throws IOException
    {
	super(null);

	// No images
	if (images.length == 0) { throw new IOException("No images"); }

	// Get image properties
	_imageWidth = images[0].getWidth();
	_imageHeight = images[0].getHeight();
	_colorModels = new ColorModel[images.length];
	_sampleModels = new SampleModel[images.length];

	// Create the temporary file
	_tempFile = File.createTempFile("debabel_", null, dir);
	_tempFile.deleteOnExit();

	// Open the file for writing
	FileImageOutputStream fios = new FileImageOutputStream(_tempFile);

	// Allocate space for the offsets
	for (int i = 0; i < images.length; i++) { fios.writeLong(0); }
	long[] offsets = new long[images.length];

	// Write the images
	long offset = 8*images.length;
	for (int i = 0; i < images.length; i++) {

	    // Save the offset
	    offsets[i] = offset;

	    // Get the image components
	    _colorModels[i] = images[i].getColorModel();
	    WritableRaster raster = images[i].getRaster();
	    _sampleModels[i] = raster.getSampleModel();
	    DataBuffer dataBuffer = raster.getDataBuffer();

	    // Write the image data to the file
	    int bytesWritten = 0;
	    if (dataBuffer instanceof DataBufferByte) {
		byte[] data = ((DataBufferByte)dataBuffer).getData();
		fios.write(data);
		bytesWritten = data.length;
	    }
	    else if (dataBuffer instanceof DataBufferShort) {
		short[] data = ((DataBufferShort)dataBuffer).getData();
		fios.writeShorts(data, 0, data.length);
		bytesWritten = 2*data.length;
	    }
	    else if (dataBuffer instanceof DataBufferUShort) {
		short[] data = ((DataBufferUShort)dataBuffer).getData();
		fios.writeShorts(data, 0, data.length);
		bytesWritten = 2*data.length;
	    }
	    else if (dataBuffer instanceof DataBufferInt) {
		int[] data = ((DataBufferInt)dataBuffer).getData();
		fios.writeInts(data, 0, data.length);
		bytesWritten = 4*data.length;
	    }
	    else if (dataBuffer instanceof DataBufferFloat) {
		float[] data = ((DataBufferFloat)dataBuffer).getData();
		fios.writeFloats(data, 0, data.length);
		bytesWritten = 4*data.length;
	    }
	    else if (dataBuffer instanceof DataBufferDouble) {
		double[] data = ((DataBufferDouble)dataBuffer).getData();
		fios.writeDoubles(data, 0, data.length);
		bytesWritten = 8*data.length;
	    }

	    // Update the offset
	    offset += bytesWritten;
	}

	// Back up and write the offsets
	fios.seek(0);
	for (int i = 0; i < images.length; i++) {
	    fios.writeLong( offsets[i] );
	}

	// Close the file stream
	fios.flush();
	fios.close();
    }

    /**
     * Returns the number of images, not including thumbnails, available from
     * the current input source.
     * 
     * @param allowSearch If true, the true number of images will be returned
     *                    even if a search is required.  If false, the reader
     *                    may return -1 without performing the search.
     *
     * @return The number of images or -1 if allowSearch is false and a search
     *         would be required.
     *
     * @throws IllegalStateException If the input source has not been set.
     * @throws IOException If an error occurs reading the information from the
     *                     input source.
     */
    public int getNumImages(boolean allowSearch) throws IOException
    {
	return _colorModels.length;
    }

    /**
     * Returns the width in pixels of the given image within the input source.
     *
     * @param imageIndex The index of the image to be queried.
     *
     * @return The width of the image in pixels.
     *
     * @throws IOException If an error occurs reading the width information from
     *                     the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public int getWidth(int imageIndex) throws IOException
    {
	return _imageWidth;
    }

    /**
     * Returns the height in pixels of the given image within the input source.
     *
     * @param imageIndex The index of the image to be queried.
     *
     * @return The height of the image in pixels.
     *
     * @throws IOException If an error occurs reading the height information
     *                     from the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public int getHeight(int imageIndex) throws IOException
    {
	return _imageHeight;
    }

    /**
     * Returns an Iterator containing possible image types to which the given
     * image may be decoded, in the form of ImageTypeSpecifiers.  At least one
     * legal image type will be returned.
     *
     * @param imageIndex The index of the image to be retrieved.
     *
     * @return An Iterator containing at least one ImageTypeSpecifier
     *         representing suggested image types for decoding the current given
     *         image.
     *
     * @throws IOException If an error occurs reading the format information
     *                     from the input source.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public Iterator getImageTypes(int imageIndex) throws IOException
    {
	return null;
    }

    /**
     * Returns an IIOMetadata object representing the metadata associated with
     * the input source as a whole (i.e., not associated with any particular
     * image), or null if the reader does not support reading metadata, is set
     * to ignore metadata, or if no metadata is available.
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     */
    public IIOMetadata getStreamMetadata() throws IOException
    {
	return null;
    }

    /**
     * Returns an IIOMetadata object containing metadata associated with the
     * given image, or null if the reader does not support reading metadata, is
     * set to ignore metadata, or if no metadata is available.
     *
     * @param imageIndex Index of the image whose metadata is to be retrieved. 
     *
     * @return An IIOMetadata object, or null.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public IIOMetadata getImageMetadata(int imageIndex) throws IOException
    {
	return null;
    }

    /**
     * Reads the image indexed by imageIndex and returns it as a complete
     * BufferedImage.  The supplied ImageReadParam is ignored.
     *
     * @param imageIndex The index of the image to be retrieved.
     * @param param An ImageReadParam used to control the reading process, or
     *              null.
     *
     * @return The desired portion of the image as a BufferedImage.
     *
     * @throws IOException If an error occurs during reading.
     * @throws IllegalStateException If the input source has not been set.
     * @throws IndexOutOfBoundsException If the supplied index is out of bounds.
     */
    public BufferedImage read(int imageIndex, ImageReadParam param)
	throws IOException
    {
	// Check the image index
	if (imageIndex >= _colorModels.length) {
	    String msg = "Illegal image index";
	    throw new IndexOutOfBoundsException(msg);
	}

	// Open the temporary file for reading
	FileImageInputStream fiis = new FileImageInputStream(_tempFile);

	// Read the offset
	_skipFully(fiis, 8*imageIndex);
	long offset = fiis.readLong();

	// Skip to the image data
	_skipFully(fiis, offset - 8*(imageIndex+1));

	// Read the image
	SampleModel sm = _sampleModels[imageIndex];
	int dataType = sm.getDataType();
	try {

	    // Construct the data buffer
	    DataBuffer dataBuffer = null;
	    if (dataType == DataBuffer.TYPE_BYTE) {
		byte[] data = new byte[_imageWidth*_imageHeight];
		fiis.readFully(data);
		dataBuffer = new DataBufferByte(data, data.length);
	    }
	    else if (dataType == DataBuffer.TYPE_SHORT) {
		short[] data = new short[_imageWidth*_imageHeight];
		fiis.readFully(data, 0, data.length);
		dataBuffer = new DataBufferShort(data, data.length);
	    }
	    else if (dataType == DataBuffer.TYPE_USHORT) {
		short[] data = new short[_imageWidth*_imageHeight];
		fiis.readFully(data, 0, data.length);
		dataBuffer = new DataBufferUShort(data, data.length);
	    }
	    else if (dataType == DataBuffer.TYPE_INT) {
		int[] data = new int[_imageWidth*_imageHeight];
		fiis.readFully(data, 0, data.length);
		dataBuffer = new DataBufferInt(data, data.length);
	    }
	    else if (dataType == DataBuffer.TYPE_FLOAT) {
		float[] data = new float[_imageWidth*_imageHeight];
		fiis.readFully(data, 0, data.length);
		dataBuffer = new DataBufferFloat(data, data.length);
	    }
	    else if (dataType == DataBuffer.TYPE_DOUBLE) {
		double[] data = new double[_imageWidth*_imageHeight];
		fiis.readFully(data, 0, data.length);
		dataBuffer = new DataBufferDouble(data, data.length);
	    }

	    // Create the raster
	    WritableRaster raster = Raster.createWritableRaster(sm, dataBuffer,
								null);
	    // Return the image
	    ColorModel cm = _colorModels[imageIndex];
	    return new BufferedImage(cm, raster, cm.isAlphaPremultiplied(),
				     new Hashtable());
	}
	catch (Exception e) {
	    String msg = "Unable to read image " + imageIndex;
	    IOException ioe = new IOException(msg);
	    ioe.initCause(e);
	    throw ioe;
	}
	finally {

	    // Close the input stream
	    fiis.close();
	}
    }

    /**
     * Allows any resources held by this object to be released.  It is important
     * for applications to call this method when they know they will no longer
     * be using this ImageReader.  Otherwise, the reader may continue to hold on
     * to resources indefinitely.
     */
    public void dispose()
    {
	// Delete the temporary file
	_tempFile.delete();
    }

    /**
     * Skips the stream to the offset.
     *
     * @param fiis Stream to skip.
     * @param bytesToSkip Number of bytes to skip.
     *
     * @throws IOException If an I/O error occurs.
     */
    private static void _skipFully(FileImageInputStream fiis, long bytesToSkip)
	throws IOException
    {
	long bytesSkipped = 0;
	while (bytesSkipped < bytesToSkip) {
	    bytesSkipped += fiis.skipBytes(bytesToSkip-bytesSkipped);
	}
    }
}
