/*
COPYRIGHT NOTICE
Copyright (c) 2002  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.io.File;

/**
 * Processor that parses a file name into its components.
 *
 * @version 2 June 2003
 */
public class FileNameParser extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Parses a file name into its components.";

  /** Input Node name for the file name. */
  private static String FILE_NAME = "File Name";

  /** Input Node description for the file name. */
  private static String FILE_DESC = "Name of the file.";

  /** Output Node name for the file path. */
  private static String PATH_NAME = "File Path";

  /** Output Node description for the file path. */
  private static String PATH_DESC = "Name of the path to the file.";

  /** Output Node name for the truncated file name. */
  private static String BASE_NAME = "Local File Name";

  /** Output Node description for the truncated file name. */
  private static String BASE_DESC =
  "Local name (without the path) of the file.";

  /** Output Node name for the path separator. */
  private static String SEPARATOR_NAME = "Separator";

  /** Output Node description for the path separator. */
  private static String SEPARATOR_DESC =
  "Character that separates the file path from the local file name.";

  /**
   * Constructs a File Name Parser with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public FileNameParser(String name)
    {
      super(name, DESC);

      // Create the Input Node for the file name
      Class type = String.class;
      InputNode inputNode = new InputNode(FILE_NAME, FILE_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node for the file path
      type = String.class;
      OutputNode outputNode = new OutputNode(PATH_NAME, PATH_DESC, type);
      _addOutputNode(outputNode);

      // Create the Output Node for the truncated file name
      type = String.class;
      outputNode = new OutputNode(BASE_NAME, BASE_DESC, type);
      _addOutputNode(outputNode);

      // Create the Output Node for the path separator
      type = String.class;
      outputNode = new OutputNode(SEPARATOR_NAME, SEPARATOR_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Constructs a File Name Parser using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public FileNameParser(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the file name
	String fileName = (String)_getInputValue(FILE_NAME, null);

	// Parse the file name
	String path = null;
	String baseName = null;
	String separator = File.separator;
	if (fileName != null) {

	  // Get the base name
	  File file = new File(fileName);
	  baseName = file.getName();

	  // Get the path
	  path = "";
	  int pathSize = fileName.length() - baseName.length() - 1;
	  if (pathSize > 0) {
	    path = fileName.substring(0, pathSize);
	  }
	}

	// Set the output values
	_setOutputValue(PATH_NAME, path);
	_setOutputValue(BASE_NAME, baseName);
	_setOutputValue(SEPARATOR_NAME, separator);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
