/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.lang.reflect.Array;

/**
 * Processor that produces the same value it receives.
 *
 * @version 12 May 2002
 */
public class Identity extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Produces the same value it receives.";

  /** Name of the Input and Output Nodes. */
  private static String NODE_NAME = "Value";

  /** Property for the value type. */
  private static String PROP_TYPE_NAME = "type";

  /** Property description for the value type. */
  private static String PROP_TYPE_DESC = "Type of value.";

  /** Property for the value description. */
  private static String PROP_DESC_NAME = "desc";

  /** Property description for the value description. */
  private static String PROP_DESC_DESC = "Description of the value.";

  /**
   * Constructs an Identity with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public Identity(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs an Identity using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public Identity(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the value type and description
      Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);
      String desc = (String)_getPropertyValue(PROP_DESC_NAME);

      // Create the Input Node
      InputNode inputNode = new InputNode(NODE_NAME, desc, type);
      _addInputNode(inputNode);

      // Create the Output Node
      OutputNode outputNode = new OutputNode(NODE_NAME, desc, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the input value
	Object value = _getInputValue(NODE_NAME, null);

	// Set the output value
	_setOutputValue(NODE_NAME, value);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[2];

      // Default value type
      String name = PROP_TYPE_NAME;
      String desc = PROP_TYPE_DESC;
      Object value = Object.class;
      Restriction restriction = Utilities.getClassRestriction();
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Default value description
      name = PROP_DESC_NAME;
      desc = PROP_DESC_DESC;
      value = "";
      properties[1] = new ProcessorProperty(name, desc, value);

      return properties;
    }
}
