/*
  COPYRIGHT NOTICE
  Copyright (c) 2010  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott;

import edu.ucla.loni.dataflow.translate.image.BufferedImageAccessor;
import edu.ucla.loni.dataflow.translate.image.BufferedImageReader;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import java.awt.image.BufferedImage;
import java.io.File;

/**
 * Processor that caches images in a temporary file.
 *
 * @version 12 November 2010
 */
public class ImageToAccessor extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Caches images in a temporary file.";

    /** Input Node name for the directory to write temporary files to. */
    private static String DIR_NAME = "Temporary File Directory";

    /** Input Node description for temporary file directory. */
    private static String DIR_DESC = "Directory for the temporary file.";

    /** Input Node name for the Buffered Images. */
    private static String IMAGE_NAME = "Images";

    /** Input Node description for the Buffered Images. */
    private static String IMAGE_DESC = "Images to cache.";

    /** Output Node name for the Buffered Image Accessor. */
    private static String ACCESSOR_NAME = "Accessors";

    /** Output Node description for the Buffered Image Accessor. */
    private static String ACCESSOR_DESC = "Accessors to the cached images.";

    /**
     * Constructs an Image To Accessor with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public ImageToAccessor(String name)
    {
	super(name, DESC);

	// Create the Input Node for the directory
	Class type = String.class;
	InputNode inputNode = new InputNode(DIR_NAME, DIR_DESC, type);
	_addInputNode(inputNode);

	// Create the Input Node for the images
	type = BufferedImage[].class;
	inputNode = new InputNode(IMAGE_NAME, IMAGE_DESC, type);
	_addInputNode(inputNode);

	// // Create the Output Node for the accessors
	type = BufferedImageAccessor[].class;
	OutputNode node = new OutputNode(ACCESSOR_NAME, ACCESSOR_DESC, type);
	_addOutputNode(node);
    }

    /**
     * Constructs an Image To Accessor using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public ImageToAccessor(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the temporary file directory
	    String dirName = (String)_getInputValue(DIR_NAME, null);
	    if (dirName == null) { return; }
	    File dir = new File(dirName);

	    // Get the images
	    BufferedImage[] images = (BufferedImage[])_getInputValue(IMAGE_NAME,
								     null);
	    if (images == null) { return; }

	    // Initialize the reader
	    FileCacheImageReader fciReader = new FileCacheImageReader(dir,
								      images);
	    BufferedImageReader biReader = new BufferedImageReader(fciReader);

	    // Create the accessors
	    BufferedImageAccessor[] accessors =
		new BufferedImageAccessor[images.length];
	    for (int i = 0; i < accessors.length; i++) {
		accessors[i] = new BufferedImageAccessor(biReader, i);
	    }

	    // Set the output value
	    _setOutputValue(ACCESSOR_NAME, accessors);
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }
}
