/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that converts from DICOM space to MINC space.
 *
 * @version 5 August 2003
 */
public class DicomToMincSpace extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Converts from DICOM space to MINC space.";

  /** Input Node name for the X coordinate of the row direction cosine. */
  private static String X_ROW_NAME = "Row Direction Cosine (X)";

  /** Input Node description for the X coord of the row direction cosine. */
  private static String X_ROW_DESC = "X coord of the row direction cosine.";

  /** Input Node name for the Y coordinate of the row direction cosine. */
  private static String Y_ROW_NAME = "Row Direction Cosine (Y)";

  /** Input Node description for the Y coord of the row direction cosine. */
  private static String Y_ROW_DESC = "Y coord of the row direction cosine.";

  /** Input Node name for the Z coordinate of the row direction cosine. */
  private static String Z_ROW_NAME = "Row Direction Cosine (Z)";

  /** Input Node description for the Z coord of the row direction cosine. */
  private static String Z_ROW_DESC = "Z coord of the row direction cosine.";

  /** Input Node name for the X coordinate of the column direction cosine. */
  private static String X_COL_NAME = "Column Direction Cosine (X)";

  /** Input Node description for X coord of the column direction cosine. */
  private static String X_COL_DESC = "X coord of the column direction cosine.";

  /** Input Node name for the Y coordinate of the column direction cosine. */
  private static String Y_COL_NAME = "Column Direction Cosine (Y)";

  /** Input Node description for Y coord of the column direction cosine. */
  private static String Y_COL_DESC = "Y coord of the column direction cosine.";

  /** Input Node name for the Z coordinate of the column direction cosine. */
  private static String Z_COL_NAME = "Column Direction Cosine (Z)";

  /** Input Node description for Z coord of the column direction cosine. */
  private static String Z_COL_DESC = "Z coord of the column direction cosine.";

  /** Input Node name for the X coordinate of the stack direction cosine. */
  private static String X_STK_NAME = "Stack Direction Cosine (X)";

  /** Input Node description for X coord of the stack direction cosine. */
  private static String X_STK_DESC = "X coord of the stack direction cosine.";

  /** Input Node name for the Y coordinate of the stack direction cosine. */
  private static String Y_STK_NAME = "Stack Direction Cosine (Y)";

  /** Input Node description for Y coord of the stack direction cosine. */
  private static String Y_STK_DESC = "Y coord of the stack direction cosine.";

  /** Input Node name for the Z coordinate of the stack direction cosine. */
  private static String Z_STK_NAME = "Stack Direction Cosine (Z)";

  /** Input Node description for Z coord of the stack direction cosine. */
  private static String Z_STK_DESC = "Z coord of the stack direction cosine.";

  /** Input Node name for the row spacing. */
  private static String ROW_STEP_NAME = "Row Spacing";

  /** Input Node description for the row spacing. */
  private static String ROW_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the row direction.";

  /** Input Node name for the column spacing. */
  private static String COL_STEP_NAME = "Column Spacing";

  /** Input Node description for the column spacing. */
  private static String COL_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the column direction.";

  /** Input Node name for the stack spacing. */
  private static String STK_STEP_NAME = "Stack Spacing";

  /** Input Node description for the stack spacing. */
  private static String STK_STEP_DESC = "Step size between adjacent " +
                                        "voxels in the stack direction.";

  /** Input Node name for the row start location. */
  private static String ROW_START_NAME = "Row Start Location";

  /** Input Node description for the row start location. */
  private static String ROW_START_DESC = "Starting location along the row " +
                                         "direction.";

  /** Input Node name for the column start location. */
  private static String COL_START_NAME = "Column Start Location";

  /** Input Node description for the column start location. */
  private static String COL_START_DESC = "Starting location along the column" +
                                         " direction.";

  /** Input Node name for the stack start location. */
  private static String STK_START_NAME = "Stack Start Location";

  /** Input Node description for the stack start location. */
  private static String STK_START_DESC = "Starting location along the stack " +
                                         "direction.";

  /** Output Node name for the name of the fastest varying dimension V1. */
  private static String V1_NAME = "Row Name";

  /** Output Node description for the name of the fastest varying dimension. */
  private static String V1_DESC = "MINC name for the fastest varying " +
                                  "dimension.";

  /** Output Node name for the name of the 2nd fastest varying dimension V2. */
  private static String V2_NAME = "Column Name";

  /** Output Node description for the name of the 2nd fastest dimension. */
  private static String V2_DESC = "MINC name for the 2nd fastest varying " +
                                  "dimension.";

  /** Output Node name for the name of the 3nd fastest varying dimension V3. */
  private static String V3_NAME = "Stack Name";

  /** Output Node description for the name of the 3nd fastest dimension. */
  private static String V3_DESC = "MINC name for the 3nd fastest varying " +
                                  "dimension.";

  /** Output Node name for the MINC xspace start location. */
  private static String XSPACE_START_NAME = "Xspace Start Location";

  /** Output Node description for the MINC xspace start location. */
  private static String XSPACE_START_DESC = "Starting location along the " +
                                            "xspace direction.";

  /** Output Node name for the MINC xspace step size. */
  private static String XSPACE_STEP_NAME = "Xspace Step";

  /** Output Node description for the MINC xspace step size. */
  private static String XSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the xspace direction.";

  /** Output Node name for the MINC xspace X direction cosine. */
  private static String XSPACE_X_NAME = "Xspace X Direction Cosine";

  /** Output Node description for the MINC xspace X direction cosine. */
  private static String XSPACE_X_DESC = "Direction cosine along X.";

  /** Output Node name for the MINC xspace Y direction cosine. */
  private static String XSPACE_Y_NAME = "Xspace Y Direction Cosine";

  /** Output Node description for the MINC xspace Y direction cosine. */
  private static String XSPACE_Y_DESC = "Direction cosine along Y.";

  /** Output Node name for the MINC xspace Z direction cosine. */
  private static String XSPACE_Z_NAME = "Xspace Z Direction Cosine";

  /** Output Node description for the MINC xspace Z direction cosine. */
  private static String XSPACE_Z_DESC = "Direction cosine along Z.";

  /** Output Node name for the MINC yspace start location. */
  private static String YSPACE_START_NAME = "Yspace Start Location";

  /** Output Node description for the MINC yspace start location. */
  private static String YSPACE_START_DESC = "Starting location along the " +
                                            "yspace direction.";

  /** Output Node name for the MINC yspace step size. */
  private static String YSPACE_STEP_NAME = "Yspace Step";

  /** Output Node description for the MINC yspace step size. */
  private static String YSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the yspace direction.";

  /** Output Node name for the MINC yspace X direction cosine. */
  private static String YSPACE_X_NAME = "Yspace X Direction Cosine";

  /** Output Node description for the MINC yspace X direction cosine. */
  private static String YSPACE_X_DESC = "Direction cosine along X.";

  /** Output Node name for the MINC yspace Y direction cosine. */
  private static String YSPACE_Y_NAME = "Yspace Y Direction Cosine";

  /** Output Node description for the MINC yspace Y direction cosine. */
  private static String YSPACE_Y_DESC = "Direction cosine along Y.";

  /** Output Node name for the MINC yspace Z direction cosine. */
  private static String YSPACE_Z_NAME = "Yspace Z Direction Cosine";

  /** Output Node description for the MINC yspace Z direction cosine. */
  private static String YSPACE_Z_DESC = "Direction cosine along Z.";

  /** Output Node name for the MINC zspace start location. */
  private static String ZSPACE_START_NAME = "Zspace Start Location";

  /** Output Node description for the MINC zspace start location. */
  private static String ZSPACE_START_DESC = "Starting location along the " +
                                            "zspace direction.";

  /** Output Node name for the MINC zspace step size. */
  private static String ZSPACE_STEP_NAME = "Zspace Step";

  /** Output Node description for the MINC zspace step size. */
  private static String ZSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the zspace direction.";

  /** Output Node name for the MINC zspace X direction cosine. */
  private static String ZSPACE_X_NAME = "Zspace X Direction Cosine";

  /** Output Node description for the MINC zspace X direction cosine. */
  private static String ZSPACE_X_DESC = "Direction cosine along X.";

  /** Output Node name for the MINC zspace Y direction cosine. */
  private static String ZSPACE_Y_NAME = "Zspace Y Direction Cosine";

  /** Output Node description for the MINC zspace Y direction cosine. */
  private static String ZSPACE_Y_DESC = "Direction cosine along Y.";

  /** Output Node name for the MINC zspace Z direction cosine. */
  private static String ZSPACE_Z_NAME = "Zspace Z Direction Cosine";

  /** Output Node description for the MINC zspace Z direction cosine. */
  private static String ZSPACE_Z_DESC = "Direction cosine along Z.";

  /**
   * Constructs a DICOM to MINC Space with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomToMincSpace(String name)
    {
      super(name, DESC);

      // Create the Input Nodes for the row quantities
      Class type = Double.class;
      _addInputNode( new InputNode(ROW_START_NAME, ROW_START_DESC, type) );
      _addInputNode( new InputNode(ROW_STEP_NAME, ROW_STEP_DESC, type) );

      // Create the Input Nodes for the row direction cosine
      type = Double.class;
      _addInputNode( new InputNode(X_ROW_NAME, X_ROW_DESC, type) );
      _addInputNode( new InputNode(Y_ROW_NAME, Y_ROW_DESC, type) );
      _addInputNode( new InputNode(Z_ROW_NAME, Z_ROW_DESC, type) );

      // Create the Input Nodes for the column quantities
      type = Double.class;
      _addInputNode( new InputNode(COL_START_NAME, COL_START_DESC, type) );
      _addInputNode( new InputNode(COL_STEP_NAME, COL_STEP_DESC, type) );

      // Create the Input Nodes for the column direction cosine
      type = Double.class;
      _addInputNode( new InputNode(X_COL_NAME, X_COL_DESC, type) );
      _addInputNode( new InputNode(Y_COL_NAME, Y_COL_DESC, type) );
      _addInputNode( new InputNode(Z_COL_NAME, Z_COL_DESC, type) );

      // Create the Input Nodes for the stack quantities
      type = Double.class;
      _addInputNode( new InputNode(STK_START_NAME, STK_START_DESC, type) );
      _addInputNode( new InputNode(STK_STEP_NAME, STK_STEP_DESC, type) );

      // Create the Input Nodes for the stack direction cosine
      type = Double.class;
      _addInputNode( new InputNode(X_STK_NAME, X_STK_DESC, type) );
      _addInputNode( new InputNode(Y_STK_NAME, Y_STK_DESC, type) );
      _addInputNode( new InputNode(Z_STK_NAME, Z_STK_DESC, type) );

      // Create the Output Nodes for the names of the dimensions
      type = String.class;
      _addOutputNode(new OutputNode(V1_NAME, V1_DESC, type));
      _addOutputNode(new OutputNode(V2_NAME, V2_DESC, type));
      _addOutputNode(new OutputNode(V3_NAME, V3_DESC, type));

      // Create the Output Nodes for the xspace quantities
      type = Double.class;
      _addOutputNode(new OutputNode(XSPACE_START_NAME, XSPACE_START_DESC,
				    type));
      _addOutputNode(new OutputNode(XSPACE_STEP_NAME, XSPACE_STEP_DESC, type));
      _addOutputNode(new OutputNode(XSPACE_X_NAME, XSPACE_X_DESC, type));
      _addOutputNode(new OutputNode(XSPACE_Y_NAME, XSPACE_Y_DESC, type));
      _addOutputNode(new OutputNode(XSPACE_Z_NAME, XSPACE_Z_DESC, type));

      // Create the Output Nodes for the yspace quantities
      type = Double.class;
      _addOutputNode(new OutputNode(YSPACE_START_NAME, YSPACE_START_DESC,
				    type));
      _addOutputNode(new OutputNode(YSPACE_STEP_NAME, YSPACE_STEP_DESC, type));
      _addOutputNode(new OutputNode(YSPACE_X_NAME, YSPACE_X_DESC, type));
      _addOutputNode(new OutputNode(YSPACE_Y_NAME, YSPACE_Y_DESC, type));
      _addOutputNode(new OutputNode(YSPACE_Z_NAME, YSPACE_Z_DESC, type));

      // Create the Output Nodes for the zspace quantities
      type = Double.class;
      _addOutputNode(new OutputNode(ZSPACE_START_NAME, ZSPACE_START_DESC,
				    type));
      _addOutputNode(new OutputNode(ZSPACE_STEP_NAME, ZSPACE_STEP_DESC, type));
      _addOutputNode(new OutputNode(ZSPACE_X_NAME, ZSPACE_X_DESC, type));
      _addOutputNode(new OutputNode(ZSPACE_Y_NAME, ZSPACE_Y_DESC, type));
      _addOutputNode(new OutputNode(ZSPACE_Z_NAME, ZSPACE_Z_DESC, type));
    }

  /**
   * Constructs a DICOM to MINC Space using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public DicomToMincSpace(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Gets the long description.
   *
   * @return A verbose description of what the Processor does.
   */
  public String getLongDescription()
    {
      StringBuffer b = new StringBuffer();

      b.append("DICOM STANDARDS:\n");
      b.append("\n");
      b.append("The x-y-z coordinate system on a patient is defined so\n");
      b.append("that the x axis points to the patient's left, the y axis\n");
      b.append("points to the patient's posterior, and the z axis points\n");
      b.append("to the patient's head:\n");
      b.append("\n");
      b.append("    (*+*)     ^ Z\n");
      b.append("     _|_      |\n");
      b.append("      |       | Y\n");
      b.append("     / \\      x -- > X\n");
      b.append("\n");
      b.append("The image plane is defined by two vectors.  Each vector\n");
      b.append("starts at the top left corner of the image plane.  The\n");
      b.append("first vector points along the top row of the image plane.\n");
      b.append("The second vector points along the left column of the\n");
      b.append("image plane:\n");
      b.append("\n");
      b.append("    *-----------------> 1st vector\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&& Image Plane &&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    |&&&&&&&&&&&&&&&&&\n");
      b.append("    V \n");
      b.append("      2nd vector\n");
      b.append("\n");
      b.append("The row and column vectors are given in the x-y-z\n");
      b.append("coordinate system defined above.  The magnitude of each\n");
      b.append("projection (x,y,z) is given as a cosine of the angle it\n");
      b.append("makes with that projection axis (== direction cosines);\n");
      b.append("e.g:\n");
      b.append("\n");
      b.append(" vector along x axis:\n");
      b.append("  -  0 degree angle with x axis\n");
      b.append("  - 90 degree angle with y axis\n");
      b.append("  - 90 degree angle with z axis\n");
      b.append("\n");
      b.append(" ==> vector = [cos(0), cos(90), cos(90)] = [1, 0, 0]\n");

      b.append("\n");
      b.append("\n");
      b.append("MINC STANDARDS:\n");
      b.append("\n");
      b.append("Voxel coordinates are defined as follows:\n");
      b.append("\n");
      b.append("V1 = Fastest varying dimension\n");
      b.append("V2 = Second fastest varying dimension\n");
      b.append("V3 = Third fastest varying dimension\n");
      b.append("\n");
      b.append("            ^\n");
      b.append("            V3                                  ^\n");
      b.append("            |                     *------------ V1\n");
      b.append("            |                     |\n");
      b.append("  image #2  |                     | ###########  rows ->\n");
      b.append("  image #1  |                     | ###########\n");
      b.append("  image #0  |           ^         | ###########  columns\n");
      b.append("            *---------- V1        |                 |\n");
      b.append("             \\                  ^                   V\n");
      b.append("              \\                 V2\n");
      b.append("               \\ ^\n");
      b.append("                 V2\n");
      b.append("\n");
      b.append("Each coordinate is given a name (\"xspace\", \"yspace\",\n");
      b.append("or \"zspace\") to indicate the patient's orientation.\n");
      b.append("Each name by itself does not indicate the orientation; it\n");
      b.append("is merely a label that denotes a set of directional\n");
      b.append("cosines, step size, and starting location (origin).  It is\n");
      b.append("then possible, for example, for \"xspace\" to denote the\n");
      b.append("Y direction if its directional cosines point in that\n");
      b.append("direction.  Note that if the step size is negative, then\n");
      b.append("the direction is reversed.\n");
      b.append("\n");
      b.append("MINC puts its coordinates in Talairach space so that\n");
      b.append("the x-y-z coordinate system on a patient is defined so\n");
      b.append("that the positive x axis points from the patient's left\n");
      b.append("to right, the positive y axis points from posterior to\n");
      b.append("anterior, and the positive z axis points from inferior to\n");
      b.append("superior:\n");
      b.append("\n");
      b.append("    (*+*)            ^ Z\n");
      b.append("     _|_             |\n");
      b.append("      |            Y |\n");
      b.append("     / \\      X < -- 0\n");

      return b.toString();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the DICOM row quantities
	Double rowStart = (Double)_getInputValue(ROW_START_NAME, null);
	Double rowStep = (Double)_getInputValue(ROW_STEP_NAME, null);
	Double rowX = (Double)_getInputValue(X_ROW_NAME, null);
	Double rowY = (Double)_getInputValue(Y_ROW_NAME, null);
	Double rowZ = (Double)_getInputValue(Z_ROW_NAME, null);

	if (rowX == null || rowY == null || rowZ == null) { return; }

	// Get the DICOM column quantities
	Double colStart = (Double)_getInputValue(COL_START_NAME, null);
	Double colStep = (Double)_getInputValue(COL_STEP_NAME, null);
	Double colX = (Double)_getInputValue(X_COL_NAME, null);
	Double colY = (Double)_getInputValue(Y_COL_NAME, null);
	Double colZ = (Double)_getInputValue(Z_COL_NAME, null);

	if (colX == null || colY == null || colZ == null) { return; }

	// Get the DICOM stack quantities
	Double stkStart = (Double)_getInputValue(STK_START_NAME, null);
	Double stkStep = (Double)_getInputValue(STK_STEP_NAME, null);
	Double stkX = (Double)_getInputValue(X_STK_NAME, null);
	Double stkY = (Double)_getInputValue(Y_STK_NAME, null);
	Double stkZ = (Double)_getInputValue(Z_STK_NAME, null);

	if (stkX == null || stkY == null || stkZ == null) { return; }

	// Assume positive step values if none are provided
	if (rowStep == null) { rowStep = new Double(1.0); }
	if (colStep == null) { colStep = new Double(1.0); }
	if (stkStep == null) { stkStep = new Double(1.0); }

	// Zero the starting location if no locations are provided
	if (rowStart == null) { rowStart = new Double(0.0); }
	if (colStart == null) { colStart = new Double(0.0); }
	if (stkStart == null) { stkStart = new Double(0.0); }

	// Convert the coordinates to Talairach space
	rowX = new Double( -rowX.doubleValue() );
	rowY = new Double( -rowY.doubleValue() );
	colX = new Double( -colX.doubleValue() );
	colY = new Double( -colY.doubleValue() );
	stkX = new Double( -stkX.doubleValue() );
	stkY = new Double( -stkY.doubleValue() );

	rowStart = new Double( -rowStart.doubleValue() );
	colStart = new Double( -colStart.doubleValue() );

	// Set the fastest varying dimension output values
	String name = _setMincCoordinateValues(rowX, rowY, rowZ, rowStep,
					       rowStart);
	_setOutputValue(V1_NAME, name);

	// Set the 2nd fastest varying dimension output values
	name = _setMincCoordinateValues(colX, colY, colZ, colStep, colStart);
	_setOutputValue(V2_NAME, name);

	// Set the slowest varying dimension output values
	name = _setMincCoordinateValues(stkX, stkY, stkZ, stkStep, stkStart);
	_setOutputValue(V3_NAME, name);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Sets the output values of the closest MINC coordinate.
   *
   * @param x Talairach directional cosine along X.
   * @param y Talairach directional cosine along Y.
   * @param z Talairach directional cosine along Z.
   * @param step Spacing along the Talairach coordinate.
   * @param start Start location of the Talairach coordinate.
   *
   * @return Name of the MINC coordinate whose output values were set.
   */
  private String _setMincCoordinateValues(Double x, Double y, Double z,
					  Double step, Double start)
    {
      double absX = Math.abs( x.doubleValue() );
      double absY = Math.abs( y.doubleValue() );
      double absZ = Math.abs( z.doubleValue() );

      // Xspace coordinate
      if (absX > absY && absX > absZ) {

	// Reverse the coordinate if negative (but then flip the step)
	if ( x.doubleValue() < 0 ) {
	  x = new Double( -x.doubleValue() );
	  y = new Double( -y.doubleValue() );
	  z = new Double( -z.doubleValue() );
	  step = new Double( -step.doubleValue() );
	}

	// Set the output values
	_setOutputValue(XSPACE_X_NAME, x);
	_setOutputValue(XSPACE_Y_NAME, y);
	_setOutputValue(XSPACE_Z_NAME, z);
	_setOutputValue(XSPACE_START_NAME, start);
	_setOutputValue(XSPACE_STEP_NAME, step);

	// Return xspace
	return "xspace";
      }

      // Yspace coordinate
      else if (absY > absX && absY > absZ) {

	// Reverse the coordinate if negative (but then flip the step)
	if ( y.doubleValue() < 0 ) {
	  x = new Double( -x.doubleValue() );
	  y = new Double( -y.doubleValue() );
	  z = new Double( -z.doubleValue() );
	  step = new Double( -step.doubleValue() );
	}

	// Set the output values
	_setOutputValue(YSPACE_X_NAME, x);
	_setOutputValue(YSPACE_Y_NAME, y);
	_setOutputValue(YSPACE_Z_NAME, z);
	_setOutputValue(YSPACE_START_NAME, start);
	_setOutputValue(YSPACE_STEP_NAME, step);

	// Return yspace
	return "yspace";
      }

      // Zspace coordinate
      else {

	// Reverse the coordinate if negative (but then flip the step)
	if ( z.doubleValue() < 0 ) {
	  x = new Double( -x.doubleValue() );
	  y = new Double( -y.doubleValue() );
	  z = new Double( -z.doubleValue() );
	  step = new Double( -step.doubleValue() );
	}

	// Set the output values
	_setOutputValue(ZSPACE_X_NAME, x);
	_setOutputValue(ZSPACE_Y_NAME, y);
	_setOutputValue(ZSPACE_Z_NAME, z);
	_setOutputValue(ZSPACE_START_NAME, start);
	_setOutputValue(ZSPACE_STEP_NAME, step);

	// Return zspace
	return "zspace";
      }
    }
}
