/*
  COPYRIGHT NOTICE
  Copyright (c) 2006  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;

/**
 * Processor that converts from NIFTI space to DICOM space.
 *
 * @version 10 August 2006
 */
public class NiftiToDicomSpace extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Converts from NIFTI space to DICOM space.";

    /** Input Node name for the scaling factor used in the rotation matrix. */
    private static String QFAC_NAME = "Q Factor";

    /** Input Node description for the scaling factor used in the rot matrix. */
    private static String QFAC_DESC =
	"Scaling factor used in the rotation matrix.";

    /** Input Node name for the quaternion b parameter. */
    private static String QUATERN_B_NAME = "Quaternion b Parameter";

    /** Input Node description for the quaternion b parameter. */
    private static String QUATERN_B_DESC =
	"B parameter used in the unit quaternion (a,b,c,d) to represent a " +
	"rotation.";

    /** Input Node name for the quaternion c parameter. */
    private static String QUATERN_C_NAME = "Quaternion c Parameter";

    /** Input Node description for the quaternion c parameter. */
    private static String QUATERN_C_DESC =
	"C parameter used in the unit quaternion (a,b,c,d) to represent a " +
	"rotation.";

    /** Input Node name for the quaternion d parameter. */
    private static String QUATERN_D_NAME = "Quaternion d Parameter";

    /** Input Node description for the quaternion d parameter. */
    private static String QUATERN_D_DESC =
	"D parameter used in the unit quaternion (a,b,c,d) to represent a " +
	"rotation.";

    /** Input Node name for the X offset added after rotation. */
    private static String QOFFSET_X_NAME = "Q Offset X";

    /** Input Node description for the X offset added after rotation. */
    private static String QOFFSET_X_DESC =
	"Offset added to the X dimension after applying the quaternion " +
	"rotation.";

    /** Input Node name for the Y offset added after rotation. */
    private static String QOFFSET_Y_NAME = "Q Offset Y";

    /** Input Node description for the Y offset added after rotation. */
    private static String QOFFSET_Y_DESC =
	"Offset added to the Y dimension after applying the quaternion " +
	"rotation.";

    /** Input Node name for the Z offset added after rotation. */
    private static String QOFFSET_Z_NAME = "Q Offset Z";

    /** Input Node description for the Z offset added after rotation. */
    private static String QOFFSET_Z_DESC =
	"Offset added to the Z dimension after applying the quaternion " +
	"rotation.";

    /** Output Node name for the X coordinate of the row direction cosine. */
    private static String X_ROW_NAME = "Row Direction Cosine (X)";

    /** Output Node description for the X coord of the row direction cosine. */
    private static String X_ROW_DESC = "X coord of the row direction cosine.";

    /** Output Node name for the Y coordinate of the row direction cosine. */
    private static String Y_ROW_NAME = "Row Direction Cosine (Y)";

    /** Output Node description for the Y coord of the row direction cosine. */
    private static String Y_ROW_DESC = "Y coord of the row direction cosine.";

    /** Output Node name for the Z coordinate of the row direction cosine. */
    private static String Z_ROW_NAME = "Row Direction Cosine (Z)";

    /** Output Node description for the Z coord of the row direction cosine. */
    private static String Z_ROW_DESC = "Z coord of the row direction cosine.";

    /** Output Node name for the X coordinate of the column direction cosine. */
    private static String X_COL_NAME = "Column Direction Cosine (X)";

    /** Output Node description for X coord of the column direction cosine. */
    private static String X_COL_DESC = "X coord of the column direction " +
	"cosine.";

    /** Output Node name for the Y coordinate of the column direction cosine. */
    private static String Y_COL_NAME = "Column Direction Cosine (Y)";

    /** Output Node description for Y coord of the column direction cosine. */
    private static String Y_COL_DESC = "Y coord of the column direction " +
	"cosine.";

    /** Output Node name for the Z coordinate of the column direction cosine. */
    private static String Z_COL_NAME = "Column Direction Cosine (Z)";

    /** Output Node description for Z coord of the column direction cosine. */
    private static String Z_COL_DESC = "Z coord of the column direction " +
	"cosine.";

    /** Output Node name for the X coordinate of the stack direction cosine. */
    private static String X_STK_NAME = "Stack Direction Cosine (X)";

    /** Output Node description for X coord of the stack direction cosine. */
    private static String X_STK_DESC = "X coord of the stack direction cosine.";

    /** Output Node name for the Y coordinate of the stack direction cosine. */
    private static String Y_STK_NAME = "Stack Direction Cosine (Y)";

    /** Output Node description for Y coord of the stack direction cosine. */
    private static String Y_STK_DESC = "Y coord of the stack direction cosine.";

    /** Output Node name for the Z coordinate of the stack direction cosine. */
    private static String Z_STK_NAME = "Stack Direction Cosine (Z)";

    /** Output Node description for Z coord of the stack direction cosine. */
    private static String Z_STK_DESC = "Z coord of the stack direction cosine.";

    /** Output Node name for the X coord of the image position. */
    private static String X_IMG_NAME = "Image Position (Patient) X";

    /** Output Node description for the X coord of the image position. */
    private static String X_IMG_DESC = "X coord of the image position.";

    /** Output Node name for the Y coord of the image position. */
    private static String Y_IMG_NAME = "Image Position (Patient) Y";

    /** Output Node description for the Y coord of the image position. */
    private static String Y_IMG_DESC = "Y coord of the image position.";

    /** Output Node name for the Z coord of the image position. */
    private static String Z_IMG_NAME = "Image Position (Patient) Z";

    /** Output Node description for the Z coord of the image position. */
    private static String Z_IMG_DESC = "Z coord of the image position.";

    /**
     * Constructs a NIFTI to DICOM Space with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public NiftiToDicomSpace(String name)
    {
	super(name, DESC);

	// Create the Input Node for the Q scaling
	Class type = Float.class;
	_addInputNode(new InputNode(QFAC_NAME, QFAC_DESC, type) );

	// Create the Input Nodes for the quaternion parameters
	type = Float.class;
	_addInputNode(new InputNode(QUATERN_B_NAME, QUATERN_B_DESC, type) );
	_addInputNode(new InputNode(QUATERN_C_NAME, QUATERN_C_DESC, type) );
	_addInputNode(new InputNode(QUATERN_D_NAME, QUATERN_D_DESC, type) );

	// Create the Input Nodes for the Q offsets
	type = Float.class;
	_addInputNode(new InputNode(QOFFSET_X_NAME, QOFFSET_X_DESC, type) );
	_addInputNode(new InputNode(QOFFSET_Y_NAME, QOFFSET_Y_DESC, type) );
	_addInputNode(new InputNode(QOFFSET_Z_NAME, QOFFSET_Z_DESC, type) );

	// Create the Output Nodes for the row direction cosine
	type = Double.class;
	_addOutputNode( new OutputNode(X_ROW_NAME, X_ROW_DESC, type) );
	_addOutputNode( new OutputNode(Y_ROW_NAME, Y_ROW_DESC, type) );
	_addOutputNode( new OutputNode(Z_ROW_NAME, Z_ROW_DESC, type) );

	// Create the Output Nodes for the column direction cosine
	type = Double.class;
	_addOutputNode( new OutputNode(X_COL_NAME, X_COL_DESC, type) );
	_addOutputNode( new OutputNode(Y_COL_NAME, Y_COL_DESC, type) );
	_addOutputNode( new OutputNode(Z_COL_NAME, Z_COL_DESC, type) );

	// Create the Output Nodes for the stack direction cosine
	type = Double.class;
	_addOutputNode( new OutputNode(X_STK_NAME, X_STK_DESC, type) );
	_addOutputNode( new OutputNode(Y_STK_NAME, Y_STK_DESC, type) );
	_addOutputNode( new OutputNode(Z_STK_NAME, Z_STK_DESC, type) );

	// Create the Output Nodes for the image position
	type = Double.class;
	_addOutputNode( new OutputNode(X_IMG_NAME, X_IMG_DESC, type) );
	_addOutputNode( new OutputNode(Y_IMG_NAME, Y_IMG_DESC, type) );
	_addOutputNode( new OutputNode(Z_IMG_NAME, Z_IMG_DESC, type) );
    }

    /**
     * Constructs a NIFTI to DICOM Space using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public NiftiToDicomSpace(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Gets the long description.
     *
     * @return A verbose description of what the Processor does.
     */
    public String getLongDescription()
    {
	StringBuffer b = new StringBuffer();

	b.append("NIFTI STANDARDS:\n");
	b.append("\n");
	b.append("The x-y-z coordinate system on a patient is defined so\n");
	b.append("that the x axis points to the patient's right, the y axis\n");
	b.append("points to the patient's anterior, and the z axis points\n");
	b.append("to the patient's head:\n");
	b.append("\n");
	b.append("    (*+*)            ^ Z\n");
	b.append("     _|_             |\n");
	b.append("      |            Y |\n");
	b.append("     / \\      X < -- 0\n");
	b.append("\n");
	b.append("When quaternions are used to define the coordinate system\n");
	b.append("(qform_code > 0), the formula for computing the\n");
	b.append("coordinate (x,y,z) from the pixel at indices (i,j,k) is:\n");
	b.append("\n");
	b.append("[x]   [R11 R12 R13] [       pixdim[1] * i]   [qoffset_x]\n");
	b.append("[y] = [R21 R22 R23] [       pixdim[2] * j] + [qoffset_y]\n");
	b.append("[z]   [R31 R32 R33] [qfac * pixdim[3] * k]   [qoffset_z]\n");
	b.append("\n");
	b.append("where the element Rij of the rotation matrix are defined\n");
	b.append("using the quaternion parameters (b,c,d).");

	b.append("\n");
	b.append("\n");
	b.append("DICOM STANDARDS:\n");
	b.append("\n");
	b.append("The x-y-z coordinate system on a patient is defined so\n");
	b.append("that the x axis points to the patient's left, the y axis\n");
	b.append("points to the patient's posterior, and the z axis points\n");
	b.append("to the patient's head:\n");
	b.append("\n");
	b.append("    (*+*)     ^ Z\n");
	b.append("     _|_      |\n");
	b.append("      |       | Y\n");
	b.append("     / \\      x -- > X\n");
	b.append("\n");
	b.append("The image plane is defined by two vectors.  Each vector\n");
	b.append("starts at the top left corner of the image plane.  The\n");
	b.append("first vector points along the top row of the image plane.\n");
	b.append("The second vector points along the left column of the\n");
	b.append("image plane:\n");
	b.append("\n");
	b.append("    *-----------------> 1st vector\n");
	b.append("    |&&&&&&&&&&&&&&&&&\n");
	b.append("    |&&&&&&&&&&&&&&&&&\n");
	b.append("    |&& Image Plane &&\n");
	b.append("    |&&&&&&&&&&&&&&&&&\n");
	b.append("    |&&&&&&&&&&&&&&&&&\n");
	b.append("    V \n");
	b.append("      2nd vector\n");
	b.append("\n");
	b.append("The row and column vectors are given in the x-y-z\n");
	b.append("coordinate system defined above.  The magnitude of each\n");
	b.append("projection (x,y,z) is given as a cosine of the angle it\n");
	b.append("makes with that projection axis (== direction cosines);\n");
	b.append("e.g:\n");
	b.append("\n");
	b.append(" vector along x axis:\n");
	b.append("  -  0 degree angle with x axis\n");
	b.append("  - 90 degree angle with y axis\n");
	b.append("  - 90 degree angle with z axis\n");
	b.append("\n");
	b.append(" ==> vector = [cos(0), cos(90), cos(90)] = [1, 0, 0]\n");

	return b.toString();
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the Q scaling factor
	    Float qfac = (Float)_getInputValue(QFAC_NAME, null);
	    if (qfac == null) { return; }

	    // Get the quaternion parameters
	    Float quatB = (Float)_getInputValue(QUATERN_B_NAME, null);
	    Float quatC = (Float)_getInputValue(QUATERN_C_NAME, null);
	    Float quatD = (Float)_getInputValue(QUATERN_D_NAME, null);

	    if (quatB == null || quatC == null || quatD == null) { return; }

	    // Get the offsets
	    Float qoffX = (Float)_getInputValue(QOFFSET_X_NAME, null);
	    Float qoffY = (Float)_getInputValue(QOFFSET_Y_NAME, null);
	    Float qoffZ = (Float)_getInputValue(QOFFSET_Z_NAME, null);

	    if (qoffX == null || qoffY == null || qoffZ == null) { return; }

	    // Calculate the quaternion parameters
	    double b = quatB.doubleValue();
	    double c = quatC.doubleValue();
	    double d = quatD.doubleValue();

	    double arg = 1 - b*b - c*c - d*d;
	    if (arg < 0) { arg = 0; }
	    double a = Math.sqrt(arg);

	    // Compute the elements of the rotation matrix
	    double r11 = a*a + b*b - c*c -d*d;
	    double r12 = 2*b*c - 2*a*d;
	    double r13 = 2*b*d + 2*a*c;

	    double r21 = 2*b*c + 2*a*d;
	    double r22 = a*a + c*c - b*b - d*d;
	    double r23 = 2*c*d - 2*a*b;

	    double r31 = 2*b*d - 2*a*c;
	    double r32 = 2*c*d + 2*a*b;
	    double r33 = a*a + d*d - c*c - b*b;

	    // Use the Q scaling factor to determine left handedness
	    if (qfac.floatValue() < 0) {
		r13 *= -1;
		r23 *= -1;
		r33 *= -1;
	    }

	    // Set the row directional cosines (negate x and y components)
	    _setOutputValue(X_ROW_NAME, new Double(-r11));
	    _setOutputValue(Y_ROW_NAME, new Double(-r21));
	    _setOutputValue(Z_ROW_NAME, new Double(r31));

	    // Set the column directional cosines (negate x and y components)
	    _setOutputValue(X_COL_NAME, new Double(-r12));
	    _setOutputValue(Y_COL_NAME, new Double(-r22));
	    _setOutputValue(Z_COL_NAME, new Double(r32));

	    // Set the stack directional cosines (negate x and y components)
	    _setOutputValue(X_STK_NAME, new Double(-r13));
	    _setOutputValue(Y_STK_NAME, new Double(-r23));
	    _setOutputValue(Z_STK_NAME, new Double(r33));

	    // Set the image position (negate x and y components)
	    _setOutputValue(X_IMG_NAME, new Double(-qoffX.doubleValue()));
	    _setOutputValue(Y_IMG_NAME, new Double(-qoffY.doubleValue()));
	    _setOutputValue(Z_IMG_NAME, new Double(qoffZ.doubleValue()));
	}

	catch (Exception e) {
	    String msg = "Unable to generate output values for \"" + this +
		"\".";
	    throw new ProcessingException(msg, e);
	}
    }
}
