/*
  COPYRIGHT NOTICE
  Copyright (c) 2009  Scott C. Neu and Arthur W. Toga

  See README.license for license notices.
*/

package edu.ucla.loni.debabel.processors.scott.dicom;

import edu.ucla.loni.Conversions;
import edu.ucla.loni.dataflow.MetadataTree;
import edu.ucla.loni.dataflow.MetadataTreeArray;
import edu.ucla.loni.dataflow.MetadataTreeInputNode;
import edu.ucla.loni.dataflow.TreeArrayOutputNode;
import edu.ucla.loni.dataflow.interpret.Interpreter;
import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.DuplicateElementException;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.jdicom.DataSet;
import edu.ucla.loni.jdicom.decoder.sv10.SV10DataSetDecoder;
import edu.ucla.loni.jdicom.plugin.DicomMetadataConversions;
import org.w3c.dom.Node;

/**
 * Processor that decodes the Siemens SV10 (0029,1010) private DICOM tag.
 *
 * @version 10 November 2009
 */
public class Siemens00291010Decoder extends Processor
{
    /** Description of this Processor. */
    private static String DESC = "Decodes the Siemens (0029, 1010) DICOM tag.";

    /** Input Node name for the value of the encoded DICOM tag. */
    private static String TAG_NAME = "DICOM Tag Value";

    /** Input Node description for the value of the encoded DICOM tag. */
    private static String TAG_DESC = "Value of the (0029,1010) DICOM tag.";

    /** Output Node name for the decoded Metadata Tree. */
    private static String TREE_NAME = "Decoded Values";

    /** Output Node description for the decoded Metadata Tree. */
    private static String TREE_DESC = "Decoded values in the (0029,1010) tag.";

    /**
     * Constructs a Siemens 0029,1010 Decoder with default Properties.
     *
     * @param name Name of the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public Siemens00291010Decoder(String name)
    {
	super(name, DESC);

	// Create the Input Node
	Class type = String.class;
	_addInputNode( new InputNode(TAG_NAME, TAG_DESC, type) );

	// Create the Output Node
	_addOutputNode( new TreeArrayOutputNode(TREE_NAME, TREE_DESC,
						_getSV10Tree()) );
    }

    /**
     * Constructs a Siemens 0029,1010 Decoder using the specified Properties.
     *
     * @param name Name of the Processor.
     * @param properties Properties for the Processor.
     *
     * @return IllegalStringException If the name contains illegal characters.
     */
    public Siemens00291010Decoder(String name, ProcessorProperty[] properties)
    {
	this(name);
    }

    /**
     * Generates values for the Output Nodes using the current values of the
     * Input Nodes.
     *
     * @throws ProcessingException If an error occurs during processing.
     */
    public void generateOutputValues() throws ProcessingException
    {
	try {

	    // Get the encoded SV10 hexadecimal string
	    String hexString = (String)_getInputValue(TAG_NAME, null);
	    if (hexString == null) { return; }

	    // Decode a Data Set from the SV10 hexadecimal string
	    byte[] bytes = Conversions.toByteArray(hexString);
	    SV10DataSetDecoder decoder = new SV10DataSetDecoder(bytes);
	    DataSet dataSet = decoder.getDataSet();

	    // Convert the Data Set to a DOM Node
	    Node sv10Node = DicomMetadataConversions.toTree("SV10", dataSet);

	    // Create a SV10 Metadata Tree and set its values
	    MetadataTree sv10Tree = _getSV10Tree();
	    Interpreter.setValues(sv10Node, sv10Tree);

	    // Set the output value
	    MetadataTree[] trees = new MetadataTree[1];
	    trees[0] = sv10Tree;
	    _setOutputValue(TREE_NAME, trees);
	}

	catch (Exception e) {}
    }

    /**
     * Gets the Siemens SV10 0029,1010 Metadata Tree.
     *
     * @return Siemens SV10 0029,1010 Metadata Tree.
     */
    private static MetadataTree _getSV10Tree()
    {
	MetadataTree sv10Tree = new MetadataTree("SV10", "SV 10");

	try {

	    // Add the Data Set tree
	    MetadataTree dataSetTree = new MetadataTree("DATA_SET", "Data Set");
	    sv10Tree.addModule(dataSetTree);

	    // Add to the Data Set tree
	    dataSetTree.addModule( _createTree("AcquisitionMatrixText",
					       "Acquisition Matrix Text",
					       "SH") );

	    dataSetTree.addModule( _createTree("Actual3DImaPartNumber",
					       "Actual 3D Ima Part Number",
					       "IS") );

	    dataSetTree.addModule( _createTree("B_value",
					       "B value",
					       "IS") );

	    dataSetTree.addModule( _createTree("Comp_AdjustedParam",
					       "Comp Adjusted Param",
					       "LT") );

	    dataSetTree.addModule( _createTree("Comp_Algorithm",
					       "Comp Algorithm",
					       "IS") );

	    dataSetTree.addModule( _createTree("Comp_AutoParam",
					       "Comp Auto Param",
					       "LT") );

	    dataSetTree.addModule( _createTree("Comp_Blended",
					       "Comp Blended",
					       "IS") );

	    dataSetTree.addModule( _createTree("Comp_JobID",
					       "Comp Job ID",
					       "LT") );

	    dataSetTree.addModule( _createTree("Comp_ManualAdjusted",
					       "Comp Manual Adjusted",
					       "IS") );

	    dataSetTree.addModule( _createTree("DiffusionDirectionality",
					       "Diffusion Directionality",
					       "CS") );

	    dataSetTree.addModule( _createTree("DiffusionGradientDirection",
					       "Diffusion Gradient Direction",
					       "FD") );

	    dataSetTree.addModule( _createTree("EchoColumnPosition",
					       "Echo Column Position",
					       "IS") );

	    dataSetTree.addModule( _createTree("EchoLinePosition",
					       "Echo Line Position",
					       "IS") );

	    dataSetTree.addModule( _createTree("EchoPartitionPosition",
					       "Echo Partition Position",
					       "IS") );

	    dataSetTree.addModule( _createTree("EchoTime",
					       "Echo Time",
					       "DS") );

	    dataSetTree.addModule( _createTree("FMRIStimulInfo",
					       "FMRI Stimul Info",
					       "IS") );

	    dataSetTree.addModule( _createTree("Filter1",
					       "Filter 1",
					       "IS") );

	    dataSetTree.addModule( _createTree("Filter2",
					       "Filter 2",
					       "IS") );

	    dataSetTree.addModule( _createTree("FlipAngle",
					       "Flip Angle",
					       "DS") );

	    dataSetTree.addModule( _createTree("FlowEncodingDirection",
					       "Flow Encoding Direction",
					       "IS") );

	    dataSetTree.addModule( _createTree("FlowEncodingDirectionString",
					       "Flow Encoding Direction String",
					       "SH") );

	    dataSetTree.addModule( _createTree("FlowVenc",
					       "Flow Venc",
					       "FD") );

	    dataSetTree.addModule( _createTree("ICE_Dims",
					       "ICE Dims",
					       "LO") );

	    dataSetTree.addModule( _createTree("ImageGroup",
					       "Image Group",
					       "US") );

	    dataSetTree.addModule( _createTree("ImageOrientationPatient",
					       "Image Orientation Patient",
					       "DS") );

	    dataSetTree.addModule( _createTree("ImagePositionPatient",
					       "Image Position Patient",
					       "DS") );

	    dataSetTree.addModule( _createTree("MeasuredFourierLines",
					       "Measured Fourier Lines",
					       "IS") );

	    dataSetTree.addModule( _createTree("NumberOfAverages",
					       "Number Of Averages",
					       "DS") );

	    dataSetTree.addModule( _createTree("NumberOfImagesInMosaic",
					       "Number Of Images In Mosaic",
					       "US") );

	    dataSetTree.addModule( _createTree("PhaseEncodingDirectionPositive",
					    "Phase Encoding Direction Positive",
					       "IS") );

	    dataSetTree.addModule( _createTree("PixelFile",
					       "Pixel File",
					       "UN") );

	    dataSetTree.addModule( _createTree("PixelFileName",
					       "Pixel File Name",
					       "UN") );

	    dataSetTree.addModule( _createTree("PixelSpacing",
					       "Pixel Spacing",
					       "DS") );

	    dataSetTree.addModule( _createTree("ProtocolSliceNumber",
					       "Protocol Slice Number",
					       "IS") );

	    dataSetTree.addModule( _createTree("RealDwellTime",
					       "Real Dwell Time",
					       "IS") );

	    dataSetTree.addModule( _createTree("RepetitionTime",
					       "Repetition Time",
					       "DS") );

	    dataSetTree.addModule( _createTree("RepetitionTimeEffective",
					       "Repetition Time Effective",
					       "DS") );

	    dataSetTree.addModule( _createTree("SequenceMask",
					       "Sequence Mask",
					       "UL") );

	    dataSetTree.addModule( _createTree("SequenceName",
					       "Sequence Name",
					       "SH") );

	    dataSetTree.addModule( _createTree("SliceLocation",
					       "Slice Location",
					       "DS") );

	    dataSetTree.addModule( _createTree("SliceMeasurementDuration",
					       "Slice Measurement Duration",
					       "DS") );

	    dataSetTree.addModule( _createTree("SliceNormalVector",
					       "Slice Normal Vector",
					       "FD") );

	    dataSetTree.addModule( _createTree("SliceThickness",
					       "Slice Thickness",
					       "DS") );

	    dataSetTree.addModule( _createTree("SpectrumTextRegionLabel",
					       "Spectrum Text Region Label",
					       "SH") );

	    dataSetTree.addModule( _createTree("TimeAfterStart",
					       "Time After Start",
					       "DS") );

	    dataSetTree.addModule( _createTree("UsedChannelMask",
					       "Used Channel Mask",
					       "UL") );

	    dataSetTree.addModule( _createTree("VoxelInPlaneRot",
					       "Voxel In Plane Rot",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelNormalCor",
					       "Voxel Normal Cor",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelNormalSag",
					       "Voxel Normal Sag",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelNormalTra",
					       "Voxel Normal Tra",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelPhaseFOV",
					       "Voxel Phase FOV",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelPositionCor",
					       "Voxel Position Cor",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelPositionSag",
					       "Voxel Position Sag",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelPositionTra",
					       "Voxel Position Tra",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelReadoutFOV",
					       "Voxel Readout FOV",
					       "DS") );

	    dataSetTree.addModule( _createTree("VoxelThickness",
					       "Voxel Thickness",
					       "DS") );
	}

	// Will not happen
	catch (Exception e) { e.printStackTrace(); }

	// Return the SV10 Metadata Tree
	return sv10Tree;
    }

    /**
     * Creates a Metadata Tree.
     *
     * @param elemName Name of the SV 10 element.
     * @param elemDesc Description of the SV 10 element.
     * @param vr Value representation of the SV 10 element.
     *
     * @return Metadata Tree.
     */
    private static MetadataTree _createTree(String elemName, String elemDesc,
					    String vr)
	throws DuplicateElementException
    {
	MetadataTree tree = new MetadataTree(elemName, elemDesc);

	// Add an Input Node for the VR
	String name = "vr";
	String desc = "Value Representation.";
	Class type = MetadataTreeInputNode.STRING_VALUE_TYPE;
	tree.addInputNode( new MetadataTreeInputNode(name, desc, type) );

	// Create the VALUE tree
	name = "VALUE";
	desc = "Values.";
	MetadataTree valueTree = new MetadataTree(name, desc);

	// Add an Input Node for the value
	name = "value";
	desc = "Value.";
	type = MetadataTreeInputNode.STRING_VALUE_TYPE;
	valueTree.addInputNode( new MetadataTreeInputNode(name, desc, type) );

	// Add an array of VALUE trees
	tree.addModule( new MetadataTreeArray(valueTree) );

	// Return the tree
	return tree;
    }
}
