/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.scott.talairach;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that converts from MINC space to Talairach space.
 *
 * @version 30 July 2003
 */
public class MincToTalairachSpace extends Processor
{
  /** Allowed Talairach space directions. */
  private static String DIRS = Talairach.POSITIVE_X + ", " + 
                               Talairach.NEGATIVE_X + ", " +
                               Talairach.POSITIVE_Y + ", " +
                               Talairach.NEGATIVE_Y + ", " +
                               Talairach.POSITIVE_Z + ", " +
                               Talairach.NEGATIVE_Z;

  /** description of this Processor. */
  private static String DESC = "Converts from MINC coordinates to Talairach " +
                               "space.";

  /** Input Node name for the name of the fastest varying dimension V1. */
  private static String V1_NAME = "Row Name";

  /** Input Node description for the name of the fastest varying dimension. */
  private static String V1_DESC = "MINC name for the fastest varying " +
                                  "dimension.";

  /** Input Node name for the name of the 2nd fastest varying dimension V2. */
  private static String V2_NAME = "Column Name";

  /** Input Node description for the name of the 2nd fastest dimension. */
  private static String V2_DESC = "MINC name for the 2nd fastest varying " +
                                  "dimension.";

  /** Input Node name for the name of the 3nd fastest varying dimension V3. */
  private static String V3_NAME = "Stack Name";

  /** Input Node description for the name of the 3nd fastest dimension. */
  private static String V3_DESC = "MINC name for the 3nd fastest varying " +
                                  "dimension.";

  /** Input Node name for the MINC xspace step size. */
  private static String XSPACE_STEP_NAME = "Xspace Step";

  /** Input Node description for the MINC xspace step size. */
  private static String XSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the xspace direction.";

  /** Input Node name for the MINC xspace X direction cosine. */
  private static String XSPACE_X_NAME = "Xspace X Direction Cosine";

  /** Input Node description for the MINC xspace X direction cosine. */
  private static String XSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC xspace Y direction cosine. */
  private static String XSPACE_Y_NAME = "Xspace Y Direction Cosine";

  /** Input Node description for the MINC xspace Y direction cosine. */
  private static String XSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC xspace Z direction cosine. */
  private static String XSPACE_Z_NAME = "Xspace Z Direction Cosine";

  /** Input Node description for the MINC xspace Z direction cosine. */
  private static String XSPACE_Z_DESC = "Direction cosine along Z.";

  /** Input Node name for the MINC yspace step size. */
  private static String YSPACE_STEP_NAME = "Yspace Step";

  /** Input Node description for the MINC yspace step size. */
  private static String YSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the yspace direction.";

  /** Input Node name for the MINC yspace X direction cosine. */
  private static String YSPACE_X_NAME = "Yspace X Direction Cosine";

  /** Input Node description for the MINC yspace X direction cosine. */
  private static String YSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC yspace Y direction cosine. */
  private static String YSPACE_Y_NAME = "Yspace Y Direction Cosine";

  /** Input Node description for the MINC yspace Y direction cosine. */
  private static String YSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC yspace Z direction cosine. */
  private static String YSPACE_Z_NAME = "Yspace Z Direction Cosine";

  /** Input Node description for the MINC yspace Z direction cosine. */
  private static String YSPACE_Z_DESC = "Direction cosine along Z.";

  /** Input Node name for the MINC zspace step size. */
  private static String ZSPACE_STEP_NAME = "Zspace Step";

  /** Input Node description for the MINC zspace step size. */
  private static String ZSPACE_STEP_DESC = "Step size between adjacent " +
                                           "voxels in the zspace direction.";

  /** Input Node name for the MINC zspace X direction cosine. */
  private static String ZSPACE_X_NAME = "Zspace X Direction Cosine";

  /** Input Node description for the MINC zspace X direction cosine. */
  private static String ZSPACE_X_DESC = "Direction cosine along X.";

  /** Input Node name for the MINC zspace Y direction cosine. */
  private static String ZSPACE_Y_NAME = "Zspace Y Direction Cosine";

  /** Input Node description for the MINC zspace Y direction cosine. */
  private static String ZSPACE_Y_DESC = "Direction cosine along Y.";

  /** Input Node name for the MINC zspace Z direction cosine. */
  private static String ZSPACE_Z_NAME = "Zspace Z Direction Cosine";

  /** Input Node description for the MINC zspace Z direction cosine. */
  private static String ZSPACE_Z_DESC = "Direction cosine along Z.";

  /** Output Node name for the Talairach coordinate of V1. */
  private static String V1_T_NAME = "Row Talairach Coordinate";

  /** Output Node description for the Talairach coordinate of V1. */
  private static String V1_T_DESC = DIRS;

  /** Output Node name for the V1 step spacing. */
  private static String V1_STEP_NAME = "Row Spacing";

  /** Output Node description for the V1 step spacing. */
  private static String V1_STEP_DESC = "Step size between adjacent " +
                                       "voxels in the rows.";

  /** Output Node name for the Talairach coordinate of V2. */
  private static String V2_T_NAME = "Column Talairach Coordinate";

  /** Output Node description for the Talairach coordinate of V2. */
  private static String V2_T_DESC = DIRS;

  /** Output Node name for the V2 step spacing. */
  private static String V2_STEP_NAME = "Column Spacing";

  /** Output Node description for the V2 step spacing. */
  private static String V2_STEP_DESC = "Step size between adjacent " +
                                       "voxels in the columns.";

  /** Output Node name for the Talairach coordinate of V3. */
  private static String V3_T_NAME = "Stack Talairach Coordinate";

  /** Output Node description for the Talairach coordinate of V3. */
  private static String V3_T_DESC = DIRS;

  /** Output Node name for the V3 step spacing. */
  private static String V3_STEP_NAME = "Stack Spacing";

  /** Output Node description for the V3 step spacing. */
  private static String V3_STEP_DESC = "Step size between adjacent " +
                                       "voxels in the stack.";

  /**
   * Constructs a MINC to Talairach Space with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public MincToTalairachSpace(String name)
    {
      super(name, DESC);

      // Create the "xspace, yspace, zspace" only Restriction
      name = "MINC Dimension Restriction";
      String desc = "Allowed values are xspace, yspace, zspace.";
      Vector options = new Vector(3);
      options.addElement("xspace");
      options.addElement("yspace");
      options.addElement("zspace");
      SetRestriction dimRestrict = new SetRestriction(name, desc, options);

      // Create the Input Nodes for the names of the dimensions
      Class type = String.class;
      InputNode inputNode = new InputNode(V1_NAME, V1_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(V2_NAME, V2_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      type = String.class;
      inputNode = new InputNode(V3_NAME, V3_DESC, type);
      inputNode.setRestriction(dimRestrict);
      _addInputNode(inputNode);

      // Create the Input Nodes for the xspace quantities
      type = Double.class;
      _addInputNode(new InputNode(XSPACE_STEP_NAME, XSPACE_STEP_DESC, type));
      _addInputNode(new InputNode(XSPACE_X_NAME, XSPACE_X_DESC, type));
      _addInputNode(new InputNode(XSPACE_Y_NAME, XSPACE_Y_DESC, type));
      _addInputNode(new InputNode(XSPACE_Z_NAME, XSPACE_Z_DESC, type));

      // Create the Input Nodes for the yspace quantities
      type = Double.class;
      _addInputNode(new InputNode(YSPACE_STEP_NAME, YSPACE_STEP_DESC, type));
      _addInputNode(new InputNode(YSPACE_X_NAME, YSPACE_X_DESC, type));
      _addInputNode(new InputNode(YSPACE_Y_NAME, YSPACE_Y_DESC, type));
      _addInputNode(new InputNode(YSPACE_Z_NAME, YSPACE_Z_DESC, type));

      // Create the Input Nodes for the zspace quantities
      type = Double.class;
      _addInputNode(new InputNode(ZSPACE_STEP_NAME, ZSPACE_STEP_DESC, type));
      _addInputNode(new InputNode(ZSPACE_X_NAME, ZSPACE_X_DESC, type));
      _addInputNode(new InputNode(ZSPACE_Y_NAME, ZSPACE_Y_DESC, type));
      _addInputNode(new InputNode(ZSPACE_Z_NAME, ZSPACE_Z_DESC, type));

      // Create the Output Nodes for the Talairach coordinates
      type = String.class;
      _addOutputNode(new OutputNode(V1_T_NAME, V1_T_DESC, type));
      _addOutputNode(new OutputNode(V2_T_NAME, V2_T_DESC, type));
      _addOutputNode(new OutputNode(V3_T_NAME, V3_T_DESC, type));

      // Create Output Nodes for the Talairach step sizes
      type = Double.class;
      _addOutputNode(new OutputNode(V1_STEP_NAME, V1_STEP_DESC, type));
      _addOutputNode(new OutputNode(V2_STEP_NAME, V2_STEP_DESC, type));
      _addOutputNode(new OutputNode(V3_STEP_NAME, V3_STEP_DESC, type));
    }

  /**
   * Constructs a MINC To Talairach Space using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @return IllegalStringException If the name contains illegal characters.
   */
  public MincToTalairachSpace(String name, ProcessorProperty[] properties)
    {
      this(name);
    }

  /**
   * Gets the long description.
   *
   * @return A verbose description of what the Processor does.
   */
  public String getLongDescription()
    {
      StringBuffer b = new StringBuffer();

      b.append("MINC STANDARDS:\n");
      b.append("\n");
      b.append("Voxel coordinates are defined as follows:\n");
      b.append("\n");
      b.append("V1 = Fastest varying dimension\n");
      b.append("V2 = Second fastest varying dimension\n");
      b.append("V3 = Third fastest varying dimension\n");
      b.append("\n");
      b.append("            ^\n");
      b.append("            V3                                  ^\n");
      b.append("            |                     *------------ V1\n");
      b.append("            |                     |\n");
      b.append("  image #2  |                     | ###########  rows ->\n");
      b.append("  image #1  |                     | ###########\n");
      b.append("  image #0  |           ^         | ###########  columns\n");
      b.append("            *---------- V1        |                 |\n");
      b.append("             \\                  ^                   V\n");
      b.append("              \\                 V2\n");
      b.append("               \\ ^\n");
      b.append("                 V2\n");
      b.append("\n");
      b.append("Each coordinate is given a name (\"xspace\", \"yspace\",\n");
      b.append("or \"zspace\") to indicate the patient's orientation.\n");
      b.append("Each name by itself does not indicate the orientation; it\n");
      b.append("is merely a label that denotes a set of directional\n");
      b.append("cosines, step size, and starting location (origin).  It is\n");
      b.append("then possible, for example, for \"xspace\" to denote the\n");
      b.append("Y direction if its directional cosines point in that\n");
      b.append("direction.  Note that if the step size is negative, then\n");
      b.append("the direction is reversed.\n");
      b.append("\n");
      b.append("MINC puts its coordinates in Talairach space so that\n");
      b.append("the x-y-z coordinate system on a patient is defined so\n");
      b.append("that the positive x axis points from the patient's left\n");
      b.append("to right, the positive y axis points from posterior to\n");
      b.append("anterior, and the positive z axis points from inferior to\n");
      b.append("superior:\n");
      b.append("\n");
      b.append("    (*+*)            ^ Z\n");
      b.append("     _|_             |\n");
      b.append("      |            Y |\n");
      b.append("     / \\      X < -- 0\n");

      return b.toString();
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      // *******
      //
      // Voxel coordinates are defined as follows:
      //
      //  V1 = Fastest varying dimension
      //  V2 = Second fastest varying dimension
      //  V3 = Third fastest varying dimension
      //
      //            ^
      //            V3                                  ^
      //            |                     *------------ V1
      //            |                     |
      //  image #2  |                     | ###########  rows ->
      //  image #1  |                     | ###########
      //  image #0  |           ^         | ###########  columns
      //            *---------- V1        |                 |
      //             \                  ^                   V
      //              \                 V2
      //               \ ^
      //                 V2
      // *******

      try {

	// Get the V1, V2, V3 dimension names
	String v1Name = (String)_getInputValue(V1_NAME, null);
	String v2Name = (String)_getInputValue(V2_NAME, null);
	String v3Name = (String)_getInputValue(V3_NAME, null);

	// Not enough information
	if (v1Name == null || v2Name == null || v3Name == null) {
	  _setOutputValues(null, null, null, null, null, null);
	  return;
	}

	// Get the xspace quantities
	Double xspaceStep = (Double)_getInputValue(XSPACE_STEP_NAME, null);
	Double xspaceX = (Double)_getInputValue(XSPACE_X_NAME, null);
	Double xspaceY = (Double)_getInputValue(XSPACE_Y_NAME, null);
	Double xspaceZ = (Double)_getInputValue(XSPACE_Z_NAME, null);

	// Use the MINC implicit coordinate system if no values
	if (xspaceX == null || xspaceY == null || xspaceZ == null) {
	  xspaceX = new Double(1.0);
	  xspaceY = new Double(0.0);
	  xspaceZ = new Double(0.0);
	}

	// Get the yspace quantities
	Double yspaceStep = (Double)_getInputValue(YSPACE_STEP_NAME, null);
	Double yspaceX = (Double)_getInputValue(YSPACE_X_NAME, null);
	Double yspaceY = (Double)_getInputValue(YSPACE_Y_NAME, null);
	Double yspaceZ = (Double)_getInputValue(YSPACE_Z_NAME, null);

	// Use the MINC implicit coordinate system if no values
	if (yspaceX == null || yspaceY == null || yspaceZ == null) {
	  yspaceX = new Double(0.0);
	  yspaceY = new Double(1.0);
	  yspaceZ = new Double(0.0);
	}

	// Get the zspace quantities
	Double zspaceStep = (Double)_getInputValue(ZSPACE_STEP_NAME, null);
	Double zspaceX = (Double)_getInputValue(ZSPACE_X_NAME, null);
	Double zspaceY = (Double)_getInputValue(ZSPACE_Y_NAME, null);
	Double zspaceZ = (Double)_getInputValue(ZSPACE_Z_NAME, null);

	// Use the MINC implicit coordinate system if no values
	if (zspaceX == null || zspaceY == null || zspaceZ == null) {
	  zspaceX = new Double(0.0);
	  zspaceY = new Double(0.0);
	  zspaceZ = new Double(1.0);
	}

	// Get the V1, V2, V3 step spacing
	Double v1Spacing = _getStepSpacing(v1Name, xspaceStep, yspaceStep,
					   zspaceStep);
	Double v2Spacing = _getStepSpacing(v2Name, xspaceStep, yspaceStep,
					   zspaceStep);
	Double v3Spacing = _getStepSpacing(v3Name, xspaceStep, yspaceStep,
					   zspaceStep);

	// Assume positive step values if none are provided
	if (xspaceStep == null) { xspaceStep = new Double(1.0); }
	if (yspaceStep == null) { yspaceStep = new Double(1.0); }
	if (zspaceStep == null) { zspaceStep = new Double(1.0); }

	// Determine the V1, V2, V3 Talairach coordinates
	String v1Direction =
	  _getTalairachCoordinate(v1Name, xspaceStep, xspaceX, xspaceY,
				  xspaceZ, yspaceStep, yspaceX, yspaceY,
				  yspaceZ, zspaceStep, zspaceX, zspaceY,
				  zspaceZ);

	String v2Direction =
	  _getTalairachCoordinate(v2Name, xspaceStep, xspaceX, xspaceY,
				  xspaceZ, yspaceStep, yspaceX, yspaceY,
				  yspaceZ, zspaceStep, zspaceX, zspaceY,
				  zspaceZ);

	String v3Direction =
	  _getTalairachCoordinate(v3Name, xspaceStep, xspaceX, xspaceY,
				  xspaceZ, yspaceStep, yspaceX, yspaceY,
				  yspaceZ, zspaceStep, zspaceX, zspaceY,
				  zspaceZ);

	// Set the output values
	_setOutputValues(v1Direction, v1Spacing, v2Direction, v2Spacing,
			 v3Direction, v3Spacing);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the step spacing for the specified MINC dimension.
   *
   * @param name Name of the dimension.
   * @param xspaceStep Xspace step size between adjacent voxels.
   * @param yspaceStep Yspace step size between adjacent voxels.
   * @param zspaceStep Zspace step size between adjacent voxels.
   *
   * @return Step spacing for the specified MINC dimension, or null if none
   *         is found.
   */
  private Double _getStepSpacing(String name, Double xspaceStep,
				 Double yspaceStep, Double zspaceStep)
    {
      Double stepSpacing = null;

      if ( name.equals("xspace") )      { stepSpacing = xspaceStep; }
      else if ( name.equals("yspace") ) { stepSpacing = yspaceStep; }
      else if ( name.equals("zspace") ) { stepSpacing = zspaceStep; }

      // Step spacing is always positive
      if (stepSpacing != null) {
	stepSpacing = new Double( Math.abs(stepSpacing.doubleValue()) );
      }

      return stepSpacing;
    }
  
  /**
   * Gets the Talairach coordinate for the specified MINC dimension.
   *
   * @param name Name of the dimension.
   * @param xspaceStep Xspace step size between adjacent voxels.
   * @param xspaceX Xspace direction cosine in the X direction.
   * @param xspaceY Xspace direction cosine in the Y direction.
   * @param xspaceZ Xspace direction cosine in the Z direction.
   * @param yspaceStep Yspace step size between adjacent voxels.
   * @param yspaceX Yspace direction cosine in the X direction.
   * @param yspaceY Yspace direction cosine in the Y direction.
   * @param yspaceZ Yspace direction cosine in the Z direction.
   * @param zspaceStep Zspace step size between adjacent voxels.
   * @param zspaceX Zspace direction cosine in the X direction.
   * @param zspaceY Zspace direction cosine in the Y direction.
   * @param zspaceZ Zspace direction cosine in the Z direction.
   *
   * @return Corresponding Talairach coordinate, or null if none is found.
   */
  private String _getTalairachCoordinate(String name, Double xspaceStep,
					 Double xspaceX, Double xspaceY,
					 Double xspaceZ, Double yspaceStep,
					 Double yspaceX, Double yspaceY,
					 Double yspaceZ, Double zspaceStep,
					 Double zspaceX, Double zspaceY,
					 Double zspaceZ)
    {
      double step = 0;
      double cosX = 0;
      double cosY = 0;
      double cosZ = 0;

      // Xspace
      if ( name.equals("xspace") ) {
	step = xspaceStep.doubleValue();
	cosX = xspaceX.doubleValue();
	cosY = xspaceY.doubleValue();
	cosZ = xspaceZ.doubleValue();
      }

      // Yspace
      else if ( name.equals("yspace") ) {
	step = yspaceStep.doubleValue();
	cosX = yspaceX.doubleValue();
	cosY = yspaceY.doubleValue();
	cosZ = yspaceZ.doubleValue();
      }

      // Zspace
      else if ( name.equals("zspace") ) {
	step = zspaceStep.doubleValue();
	cosX = zspaceX.doubleValue();
	cosY = zspaceY.doubleValue();
	cosZ = zspaceZ.doubleValue();
      }

      // Return the corresponding Talairach coordinate
      return _getTalairachCoordinate(step, cosX, cosY, cosZ);
    }

  /**
   * Gets the Talairach coordinate for the given parameters.
   *
   * @param step Step size between adjacent voxels.
   * @param cosX Direction cosine in the X direction.
   * @param cosY Direction cosine in the Y direction.
   * @param cosZ Direction cosine in the Z direction.
   *
   * @return Corresponding Talairach coordinate, or null if none is found.
   */
  private String _getTalairachCoordinate(double step, double cosX,
					 double cosY, double cosZ)
    {
      String direction = null;

      // Divide space using cones of 45 degrees around each axis
      double cos45 = 0.707106781;

      // Positive X direction
      if (cosX > cos45) {

	// No Y and Z direction
	if ( Math.abs(cosY) < cos45 && Math.abs(cosZ) < cos45 ) {
	  direction = Talairach.POSITIVE_X;
	}
      }

      // No X direction
      else if ( Math.abs(cosX) < cos45 ) {

	// Positive Y direction
	if (cosY > cos45) {

	  // No Z direction
	  if ( Math.abs(cosZ) < cos45 ) { direction = Talairach.POSITIVE_Y; }
	}

	// No Y direction
	else if ( Math.abs(cosY) < cos45 ) {

	  // Positive Z direction
	  if (cosZ > cos45) { direction = Talairach.POSITIVE_Z; }

	  // Negative Z direction
	  else if (cosZ < -1*cos45) { direction = Talairach.NEGATIVE_Z; }
	}

	// Negative Y direction
	else if (cosY < -1*cos45) {

	  // No Z direction
	  if ( Math.abs(cosZ) < cos45 ) { direction = Talairach.NEGATIVE_Y; }
	}
      }

      // Negative X direction
      else if (cosX < -1*cos45) {

	// No Y and Z direction
	if ( Math.abs(cosY) < cos45 && Math.abs(cosZ) < cos45 ) {
	  direction = Talairach.NEGATIVE_X;
	}
      }

      // Reverse the direction if the step is negative
      if (direction != null && step < 0) {
	if (direction == Talairach.POSITIVE_X) {
	  direction = Talairach.NEGATIVE_X;
	}
	else if (direction == Talairach.NEGATIVE_X) {
	  direction = Talairach.POSITIVE_X;
	}
	else if (direction == Talairach.POSITIVE_Y) {
	  direction = Talairach.NEGATIVE_Y;
	}
	else if (direction == Talairach.NEGATIVE_Y) {
	  direction = Talairach.POSITIVE_Y;
	}
	else if (direction == Talairach.POSITIVE_Z) {
	  direction = Talairach.NEGATIVE_Z;
	}
	else if (direction == Talairach.NEGATIVE_Z) {
	  direction = Talairach.POSITIVE_Z;
	}
      }

      // Return the Talairach coordinate
      return direction;
    }

  /**
   * Sets the output values to the specified values.
   *
   * @param v1Direction Talairach coordinate of V1.
   * @param v1Spacing Step spacing for V1.
   * @param v2Direction Talairach coordinate of V2.
   * @param v2Spacing Step spacing for V2.
   * @param v3Direction Talairach coordinate of V3.
   * @param v3Spacing Step spacing for V3.
   */
  private void _setOutputValues(String v1Direction, Double v1Spacing,
				String v2Direction, Double v2Spacing,
				String v3Direction, Double v3Spacing)
    {
      _setOutputValue(V1_T_NAME, v1Direction);
      _setOutputValue(V1_STEP_NAME, v1Spacing);
      _setOutputValue(V2_T_NAME, v2Direction);
      _setOutputValue(V2_STEP_NAME, v2Spacing);
      _setOutputValue(V3_T_NAME, v3Direction);
      _setOutputValue(V3_STEP_NAME, v3Spacing);
    }
}
