/*
COPYRIGHT NOTICE
Copyright (c) 2002  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.util.Vector;

/**
 * Processor that returns a default Object if no valid Object is passed to it.
 *
 * @author Keith R. Ouellette
 * 
 * @version 5 June 2003
 */
public class PrimitiveDefaulter extends Processor
{ 
  /** Description of this Processor. */
  private static String DESC = "Returns a default if there's no valid input.";
  
  /** Input Node Name for the Object to check. */
  private static String IN_NAME = "Object";

  /** Input Node Description for the Object to check. */
  private static String IN_DESC = "Object to check.";

  /** Output Node Name for the returned Object. */
  private static String OUT_NAME = "Validated Object";

  /** Output Node Description for the returned Object. */
  private static String OUT_DESC = "Original Object or the default Object.";
 
  /** Property for the default value. */
  private static String PROP_VALUE_NAME = "value";

  /** Property description for the default value. */
  private static String PROP_VALUE_DESC = "Default value.";

  /** Property for the null value type. */
  private static String PROP_TYPE_NAME = "default type";

  /** Property description for the null value type. */
  private static String PROP_TYPE_DESC = "Type of default value."; 

  /** Property for the null value type. */
  private static String PROP_MAKE_NULL_NAME = "default value null?";

  /** Property description for the null value type. */
  private static String PROP_MAKE_NULL_DESC = "Make the default value null?.";

  /**
   * Constructs a PrimitiveDefaulter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public PrimitiveDefaulter(String name)
    {
      this(name, _getDefaultProperties());
    }
 
  /**
   * Constructs a PrimitiveDefaulter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public PrimitiveDefaulter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      try { 
	// Add the Properties
	_addProperties(_getDefaultProperties(), properties);
	
	// Get the default value type and description
	Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);
	
	// Support 1-D array
	String className = type.getName();
      
	if ( type.isArray() ) {
	  // Get the Class of the array
	  type = Class.forName( className.substring(2, name.length()) );       
	  // Get a 1-D array of the Class
	  type = Array.newInstance(type, 0).getClass();
	}
      
	// Create left operand InputNode
	InputNode inputNode = new InputNode(IN_NAME, IN_DESC, type);
	_addInputNode(inputNode);
	
	// Create the Output Node for the result of the operation
	OutputNode outputNode = new OutputNode(OUT_NAME, OUT_DESC, type);
	_addOutputNode(outputNode);
      }
      catch(Exception e) {
	String msg = "Unable to create \"" + this + "\".";
	throw new IllegalArgumentException(msg);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[3];

      // Primitive classes
      Vector optionSet = new Vector();
      optionSet.add( Boolean.class );
      optionSet.add( Byte.class );
      optionSet.add( Double.class );
      optionSet.add( Float.class );
      optionSet.add( Integer.class );
      optionSet.add( Long.class );
      optionSet.add( Short.class );
      optionSet.add( String.class );

      // Create a Set Restriction
      String name = "Allowed Primitive Types";
      String desc = "Types that can be defaulted to.";
      Restriction restriction = new SetRestriction(name, desc, optionSet);

      // Default value type
      name = PROP_TYPE_NAME;
      desc = PROP_TYPE_DESC;
      Object value = String.class;
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      // Default value
      name = PROP_VALUE_NAME;
      desc = PROP_VALUE_DESC;
      value = new String();
      properties[1] = new ProcessorProperty(name, desc, value);
 
      // Make a restriction for the "make null" property
      optionSet = new Vector();
      optionSet.add( new Boolean(true) );		     
      optionSet.add( new Boolean(false) );
      name = "Allowed options";
      desc = "Allowed answers to the question the property poses.";
      restriction = new SetRestriction(name, desc, optionSet);

      // Make default value null property
      name = PROP_MAKE_NULL_NAME;
      desc = PROP_MAKE_NULL_DESC;
      value = new Boolean(false);
      properties[2] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the input Object
	Object value = _getInputValue(IN_NAME, null);
	// Get the default value type
	Class type = (Class)_getPropertyValue(PROP_TYPE_NAME);
	// Get the make default null option
	Boolean makeNull = (Boolean)_getPropertyValue(PROP_MAKE_NULL_NAME);

	if ( !type.isInstance(value) ) {
	  // Create default value

	  // Get the Constructor of the Class with one String argument
	  Class[] parameters = new Class[1];
	  parameters[0] = String.class;
	  Constructor constructor = type.getDeclaredConstructor(parameters);
	  
	  // Set a new instance of the Class
	  Object[] arguments = new Object[1];
	  arguments[0] = (String) _getPropertyValue(PROP_VALUE_NAME);

	  Object defaultValue = constructor.newInstance(arguments);
	  value = (makeNull.booleanValue() == true? null: defaultValue);
	}
	// Set the returned Object
	_setOutputValue(OUT_NAME, value);
      }
      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }
}
