/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that converts a Number or String into a different Number or String
 * type.
 *
 * @author Scott C. Neu 
 *
 * @version 3 June 2003
 */
public class PrimitiveTypeConverter extends Processor
{
  /** Description of this Processor. */
  private static String DESC =
  "Converts a Number or String into a different type.";

  /** Input Node Name for the primitive to convert. */
  private static String PRIMITIVE_NAME = "Number or String";

  /** Input Node Description for the primitive to convert. */
  private static String PRIMITIVE_DESC =
  "Number or String to convert to another type.";

  /** Output Node Name for the converted primitive. */
  private static String NEW_PRIMITIVE_NAME = "Converted Number or String";

  /** Output Node Description for the converted primitive. */
  private static String NEW_PRIMITIVE_DESC =
  "Number or String after conversion.";

  /** Property for the primitive type. */
  private static String PROP_NAME = "type";

  /** Property description for the primitive type. */
  private static String PROP_DESC = "Type of Number or String to convert to.";

  /**
   * Constructs a Primitive Type Converter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public PrimitiveTypeConverter(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a Primitive Type Converter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public PrimitiveTypeConverter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Get the primitive type
      Class primitiveType = (Class)_getPropertyValue(PROP_NAME);

      // Create the Input Node
      Class type = Object.class;
      InputNode inputNode = new InputNode(PRIMITIVE_NAME, PRIMITIVE_DESC,type);
      _addInputNode(inputNode);

      // Create the Output Node
      OutputNode node = new OutputNode(NEW_PRIMITIVE_NAME, NEW_PRIMITIVE_DESC,
				       primitiveType);
      _addOutputNode(node);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {

	// Get the primitive value
	Object value = _getInputValue(PRIMITIVE_NAME, null);

	// Convert a Number to the specified type
	Class type = (Class)_getPropertyValue(PROP_NAME);
	Object convert = null;
	if (value instanceof Number) {
	  Number n = (Number)value;

	  if (type == Byte.class) { convert = new Byte(n.byteValue()); }
	  if (type == Short.class) { convert = new Short(n.shortValue()); }
	  if (type == Integer.class) { convert = new Integer(n.intValue()); }
	  if (type == Long.class) { convert = new Long(n.longValue()); }
	  if (type == Float.class) { convert = new Float(n.floatValue()); }
	  if (type == Double.class) { convert = new Double(n.doubleValue()); }
	  if (type == String.class) { convert = n.toString(); }
	}

	// Convert a String to the specified type
	else if (value instanceof String) {
	  String s = (String)value;

	  if (type == Byte.class) { convert = Byte.valueOf(s.trim()); }
	  if (type == Short.class) { convert = Short.valueOf(s.trim()); }
	  if (type == Integer.class) { convert = Integer.valueOf(s.trim()); }
	  if (type == Long.class) { convert = Long.valueOf(s.trim()); }
	  if (type == Float.class) { convert = Float.valueOf(s.trim()); }
	  if (type == Double.class) { convert = Double.valueOf(s.trim()); }
	  if (type == String.class) { convert = s; }
	}

	// Set the conversion
	_setOutputValue(NEW_PRIMITIVE_NAME, convert);
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Primitive classes
      Vector optionSet = new Vector();
      optionSet.add( Byte.class );
      optionSet.add( Double.class );
      optionSet.add( Float.class );
      optionSet.add( Integer.class );
      optionSet.add( Long.class );
      optionSet.add( Short.class );
      optionSet.add( String.class );

      // Create a Set Restriction
      String name = "Allowed Primitive Types";
      String desc = "Types that can be converted to.";
      Restriction restriction = new SetRestriction(name, desc, optionSet);

      // Default primitive type
      name = PROP_NAME;
      desc = PROP_DESC;
      Object value = String.class;
      properties[0] = new ProcessorProperty(name, desc, value, restriction);

      return properties;
    }
}
