/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.stdlib;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector;

/**
 * Processor that compares a String to another with respect to the user 
 * specified criteria.
 *
 * @author Keith R. Ouellette 
 *
 * @version 3 June 2003
 */
public class StringComparer extends Processor
{
  /** Description of this Processor. */
  private static String DESC = "Compares a String to another with respect" + 
                               "to the user specified criteria.";

  /** Input Node name of the String to check. */
  private static String STRING_NAME = "string";

  /** Input Node description of the String to check. */
  private static String STRING_DESC = "String to check.";

  /** Input Node name of the String argument. */
  private static String ARG_NAME = "arg";

  /** Input Node description of the String argument. */
  private static String ARG_DESC = "String argument.";
  
  /** Output Node name of the Boolean returned. */
  private static String BOOLEAN_NAME = "result";

  /** Output Node description of the Boolean returned. */
  private static String BOOLEAN_DESC = 
                  "Boolean returned according to the check.";
  
  /** Property for the operation. */
  private static String OP_PROP_NAME = "operation";

  /** Property description for the operation. */
  private static String OP_PROP_DESC = "String operation to execute.";
  
  /** Comparing criteria. */
  private static String CONTAINS_SUB = "contains substring";
  
  /** Comparing criteria. */
  private static String ENDS_WITH = "ends with";

  /** Comparing criteria. */
  private static String EQUALS = "equals";

  /** Comparing criteria. */
  private static String EQUALS_IGNORE = "equals (ignore case)";

  /** Comparing criteria. */
  private static String MATCHES = "matches";

  /** Comparing criteria. */
  private static String STARTS_WITH = "starts with"; 

  /**
   * Constructs a StringComparer with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public StringComparer(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a StringComparer using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public StringComparer(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the String to check Input Node
      Class type = String.class;
      InputNode inputNode = new InputNode(STRING_NAME, STRING_DESC, type);
      _addInputNode(inputNode);
      
      // Create the String argument Input Node
      inputNode = new InputNode(ARG_NAME, ARG_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      type = Boolean.class;
      OutputNode outputNode = new OutputNode(BOOLEAN_NAME, BOOLEAN_DESC, type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	// Get the String to check
	String value = (String)_getInputValue(STRING_NAME, "");
	// Get the String argument
	String arg = (String)_getInputValue(ARG_NAME, "");
	// Get the String operation 
	String op = (String)_getPropertyValue(OP_PROP_NAME);

	Boolean bool = null;

	if(op.equals(CONTAINS_SUB)) {
	  bool = new Boolean(value.indexOf(arg) != -1);
	}
	else if (op.equals(ENDS_WITH)) {
	  bool = new Boolean(value.endsWith(arg));
	}
	else if (op.equals(EQUALS)) {
	  bool = new Boolean(value.equals(arg));
	}
	else if (op.equals(EQUALS_IGNORE)) {
	  bool = new Boolean(value.equalsIgnoreCase(arg));
	}
	else if (op.equals(MATCHES)) {
	  bool = new Boolean(value.matches(arg));
	}
	else if (op.equals(STARTS_WITH)) {
	  bool = new Boolean(value.startsWith(arg));
	}
	_setOutputValue(BOOLEAN_NAME, bool);	
      }

      catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      ProcessorProperty[] properties = new ProcessorProperty[1];

      // Default boolean operation
      String name = OP_PROP_NAME;
      String desc = OP_PROP_DESC;
      Object value = EQUALS;

      // Create the Vector of supported String operations.
      Vector ops = new Vector();
      ops.add(CONTAINS_SUB);
      ops.add(ENDS_WITH);
      ops.add(EQUALS);
      ops.add(EQUALS_IGNORE);
      ops.add(MATCHES);
      ops.add(STARTS_WITH);
    
      // Create a Restriction to limit the allowed supported operations
      String REST_NAME = "Allowable operations";
      String REST_DESC = "Allowable binary operations for this processor";
      Restriction restriction = new SetRestriction(REST_NAME, REST_DESC, ops);

      properties[0] = new ProcessorProperty(name, desc, value, restriction);
      return properties;
    }
}
