/*
COPYRIGHT NOTICE
Copyright (c) 2003  Scott C. Neu, Daniel J. Valentino, and Arthur W. Toga

See README.license for license notices.
 */

package edu.ucla.loni.debabel.processors.vann;

import edu.ucla.loni.flow.InputNode;
import edu.ucla.loni.flow.OutputNode;
import edu.ucla.loni.flow.Restriction;
import edu.ucla.loni.flow.exceptions.ProcessingException;
import edu.ucla.loni.flow.process.Processor;
import edu.ucla.loni.flow.process.ProcessorProperty;
import edu.ucla.loni.flow.restrictions.RangeRestriction;
import edu.ucla.loni.flow.restrictions.SetRestriction;
import java.util.Vector; 

/**
 * Processor that takes integer values of day, month, and year, and formats
 * it correctly for dicom date format.
 *
 * @author Bryan C. Vann
 * 
 * @version 24 June 2003
 */
public class DicomDateFormatter extends Processor
{   
  /** Description of this Processor. */
  private static String DESC = "Converts date elements to dicom format.";

  /** Base Input Node name for the year. */
  private static String YEAR_NAME = "Year"; 

  /** Input Node description of the year. */
  private static String YEAR_DESC = "Year part of the date";
				
  /** Base Input Node name for the month. */
  private static String MONTH_NAME = "Month"; 

  /** Input Node description of the month. */
  private static String MONTH_DESC = "Month part of the date";
  
  /** Base Input Node name for the date. */
  private static String DATE_NAME = "Date"; 

  /** Input Node description of the date. */
  private static String DATE_DESC = "Date part of the date";

  /** Output Node name of the returned String. */
  private static String FORMATTED_NAME = "Dicom date";

  /** Output Node description of the returned String. */
  private static String FORMATTED_DESC = "Dicom date formatted string.";

  /**
   * Constructs a DicomDateFormatter with default Properties.
   *
   * @param name Name of the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public DicomDateFormatter(String name)
    {
      this(name, _getDefaultProperties());
    }

  /**
   * Constructs a DicomDateFormatter using the specified Properties.
   *
   * @param name Name of the Processor.
   * @param properties Properties for the Processor.
   *
   * @throws IllegalStringException If the name contains illegal characters.
   */
  public DicomDateFormatter(String name, ProcessorProperty[] properties)
    {
      super(name, DESC);

      // Add the Properties
      _addProperties(_getDefaultProperties(), properties);

      // Create the Input Node
      Class type = Integer.class;
      InputNode inputNode = new InputNode(YEAR_NAME, YEAR_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node
      type = Integer.class;
      inputNode = new InputNode(MONTH_NAME, MONTH_DESC, type);
      _addInputNode(inputNode);

      // Create the Input Node
      type = Integer.class;
      inputNode = new InputNode(DATE_NAME, DATE_DESC, type);
      _addInputNode(inputNode);

      // Create the Output Node
      type = String.class;
      OutputNode outputNode = new OutputNode(FORMATTED_NAME, FORMATTED_DESC,
					     type);
      _addOutputNode(outputNode);
    }

  /**
   * Generates values for the Output Nodes using the current values of the
   * Input Nodes.
   *
   * @throws ProcessingException If an error occurs during processing.
   */
  public void generateOutputValues() throws ProcessingException
    {
      try {
	//Get the Year
	String year = 
	  ((Integer)_getInputValue(YEAR_NAME, new Integer("0"))).toString();

	//Get the Month
	String month = 
	  ((Integer)_getInputValue(MONTH_NAME, new Integer("0"))).toString();

	//Get the Date
	String date = 
	  ((Integer)_getInputValue(DATE_NAME, new Integer("0"))).toString();
	
	while(year.length() < 4){
	  year = "0" + year;
	}
	year = year.substring(0,4);

	while(month.length() < 2){
	  month = "0" + month;
	}
	month = month.substring(0,2);

	while(date.length() < 2){
	  date = "0" + date;
	}
	date = date.substring(0,2);
	
	_setOutputValue(FORMATTED_NAME, year + month + date);
      }catch (Exception e) {
	String msg = "Unable to generate output values for \"" + this + "\".";
	throw new ProcessingException(msg, e);
      }
    }	

  /**
   * Gets the default Properties.
   *
   * @return Default Processor Properties.
   */
  private static ProcessorProperty[] _getDefaultProperties()
    {
      return new ProcessorProperty[0];
    }
}



      
